# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.validator
"""
import os.path

from logging import getLogger
from os.path import exists

from ..file import File
from ..issues import Issue
from ..util import escape

from .parser import Node
from .environment import Environment
from .model import LanguageModelFactory

LOG = getLogger(__name__)

class LaTeXValidator(object):
    """
    This validates the following aspects by walking the document model
    and using the outline:

     * unused labels
     * unclosed environments, also "\[" and "\]"
    """

    def __init__(self):
        self._environment = Environment()
        #the the language_model singleton
        self._language_model = LanguageModelFactory().get_language_model()

    def validate(self, document_node, outline, issue_handler, document_preferences):
        """
        Validate a LaTeX document

        @param document_node: the root node of the document tree
        @param outline: a LaTeX outline object
        @param issue_handler: an object implementing IIssueHandler
        @param document_preferences: a DocumentPreferences object, so we can query the
               graphics file extensions, etc
        """

        LOG.debug("Validating")

        #~ # TODO: this is dangerous, the outline object could be outdated
        #~ self._outline = outline

        # cache some settings now to save time
        self._potential_graphics_extensions = [""] + document_preferences.get("graphics-extensions").split(",")
        self._potential_graphics_paths = document_preferences.get("graphics-paths").split(",")
        self._extra_issue_commands = set([c for c in document_preferences.get("extra-issue-commands").split(",") if len(c)])

        # prepare a map for checking labels
        self._labels = {}
        for label in outline.labels:
            self._labels[label.value] = [label, False]

        self._environStack = []

        self._checkRefs = True

        self._run(document_node, issue_handler)

        # evaluate label map
        # TODO this block was commented out because it is not possible to know for sure when
        #      a label was referenced. For example, using memoir class the commands fref and
        #      tref can be used instead but the label would still be highlighted as warning
        #      due to being unused. Maybe this could one day be made into an option or the
        #      check for used have been improved.
        #for label, used in self._labels.values():
        #    if not used:
        #        # FIXME: we need to know in which File the label was defined!
        #        issue_handler.issue(Issue("Label <b>%s</b> is never used" % escape(label.value), label.start, label.end, label.file, Issue.SEVERITY_WARNING))

    def _run(self, parentNode, issue_handler):
        """
        Recursive method validation
        """
        for node in parentNode:
            recurse = True
#            if node.type == Node.DOCUMENT:
#
#                self._log.debug("DOCUMENT: %s" % node.value)
#
#                # the document node contains the File object as value
#                self._file = node.value

            if node.type == Node.COMMAND:
                if node.value == "begin":
                    try:
                        # push environment on stack
                        environ = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        self._environStack.append((environ, node.start, node.lastEnd))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "end":
                    try:
                        # check environment
                        environ = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        try:
                            tEnviron, tStart, tEnd = self._environStack.pop()
                            if tEnviron != environ:
                                issue_handler.issue(Issue("Environment <b>%s</b> has to be ended before <b>%s</b>" % (escape(tEnviron), escape(environ)), node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))
                        except IndexError:
                            issue_handler.issue(Issue("Environment <b>%s</b> has no beginning" % escape(environ), node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "[":
                    # push eqn env on stack
                    self._environStack.append(("[", node.start, node.end))

                elif node.value == "]":
                    try:
                        tEnviron, tStart, tEnd = self._environStack.pop()
                        if tEnviron != "[":
                            issue_handler.issue(Issue("Environment <b>%s</b> has to be ended before <b>]</b>" % escape(tEnviron), node.start, node.end, node.file, Issue.SEVERITY_ERROR))
                    except IndexError:
                        issue_handler.issue(Issue("Environment <b>%s</b> has no beginning" % escape(environ), node.start, node.end, node.file, Issue.SEVERITY_ERROR))

                elif self._language_model.is_ref_command(node.value):
                    # mark label as used
                    try:
                        label = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        try:
                            self._labels[label][1] = True
                        except KeyError:
                            issue_handler.issue(Issue("Label <b>%s</b> has not been defined" % escape(label), node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif self._checkRefs and (node.value == "includegraphics"):
                    try:
                        # check referenced image file
                        target = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        if len(target) > 0:
                            found = False

                            if File.is_absolute(target):
                                for ext in self._potential_graphics_extensions:
                                    if exists(target + ext):
                                        found = True
                                        break
                            else:
                                for p in self._potential_graphics_paths:
                                    if found: break
                                    for ext in self._potential_graphics_extensions:
                                        if found: break
                                        filename = os.path.abspath(os.path.join(node.file.dirname, p, target) + ext)
                                        if os.path.exists(filename):
                                            found = True

                            if not found:
                                issue_handler.issue(Issue("Image <b>%s</b> could not be found" % escape(target), node.start, node.lastEnd, node.file, Issue.SEVERITY_WARNING))
                        else:
                            issue_handler.issue(Issue("No image file specified", node.start, node.lastEnd, node.file, Issue.SEVERITY_WARNING))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif self._checkRefs and (node.value == "include" or node.value == "input"):
                    # check referenced tex file
                    try:
                        target = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        if len(target) > 0:
                            if File.is_absolute(target):
                                filename = target
                            else:
                                file = File.create_from_relative_path(target, node.file.dirname)
                                filename = file.path

                            # an input may be specified without the extension
                            potential_extensions = ["", ".tex"]

                            found = False
                            for ext in potential_extensions:
                                if exists(filename + ext):
                                    found = True
                                    break

                            if not found:
                                issue_handler.issue(Issue("Document <b>%s</b> could not be found" % escape(filename), node.start, node.lastEnd, node.file, Issue.SEVERITY_WARNING))
                    except IndexError:        # firstOfType failed
                        # this happens on old-style syntax like "\input myfile"
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif self._checkRefs and node.value == "bibliography":
                    try:
                        # check referenced BibTeX file(s)
                        value = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        for target in value.split(","):
                            if File.is_absolute(target):
                                filename = target
                            else:
                                file = File.create_from_relative_path(target, node.file.dirname)
                                filename = file.path

                            # a bib file may be specified without the extension
                            potential_extensions = ["", ".bib"]

                            found = False
                            for ext in potential_extensions:
                                if exists(filename + ext):
                                    found = True
                                    break

                            if not found:
                                issue_handler.issue(Issue("Bibliography <b>%s</b> could not be found" % escape(filename), node.start, node.lastEnd, node.file, Issue.SEVERITY_WARNING))
                    except IndexError:        # firstOfType failed
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "newcommand":
                    # don't validate in newcommand definitions
                    recurse = False

                elif node.value == "bibliographystyle":
                    try:
                        # check if style exists
                        value = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText

                        # search the TeX environment
                        if not self._environment.file_exists("%s.bst" % value):
                            # search the working directory
                            bst_file = File.create_from_relative_path("%s.bst" % value, node.file.dirname)
                            if not bst_file.exists:
                                issue_handler.issue(Issue("Bibliography style <b>%s</b> could not be found" % escape(value), node.start, node.lastEnd, node.file, Issue.SEVERITY_WARNING))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value in self._extra_issue_commands:
                    try:
                        text = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                    except IndexError:
                        text = node.value
                    issue_handler.issue(Issue(text, node.start, node.lastEnd, node.file, Issue.SEVERITY_TASK))

            if recurse:
                self._run(node, issue_handler)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#               2011 Ignacio Casal Quinteiro
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
resources
"""

import logging
import os.path
import errno
from .singleton import Singleton

_log = logging.getLogger("resources")

class Resources(Singleton):
    def __init_once__(self):
        pass

    def set_dirs(self, userdir, systemdir):
        self.userdir = userdir
        self.systemdir = systemdir

        # Make sure dir exists
        try:
            os.makedirs(userdir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def get_user_dir(self):
        return self.userdir

    def get_system_dir(self):
        return self.systemdir

    def get_user_file(self, user_file):
        return os.path.join(self.userdir, user_file)

    def get_ui_file(self, ui_name):
        return os.path.join(self.systemdir, "ui", ui_name)

    def get_icon(self, icon_name):
        return os.path.join(self.systemdir, "icons", icon_name)

    def get_data_file(self, data_name):
        return os.path.join(self.systemdir, data_name)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base.views
"""

import logging

from gi.repository import Gtk, GdkPixbuf

from ..resources import Resources
from ..panelview import PanelView
from ..issues import Issue, IStructuredIssueHandler
from ..gldefs import _

LOG = logging.getLogger(__name__)

class ToolView(PanelView, IStructuredIssueHandler):
    """
    """

    def __init__(self, context, editor):
        PanelView.__init__(self, context)
        self._handlers = {}

        self._ICON_RUN = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("run.png"))
        self._ICON_FAIL = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("error.png"))
        self._ICON_SUCCESS = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("okay.png"))
        self._ICON_ERROR = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("error.png"))
        self._ICON_WARNING = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("warning.png"))
        self._ICON_ABORT = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("abort.png"))

        grid = Gtk.Grid()
        self.add(grid)

        self._scroll = Gtk.ScrolledWindow()
        self._scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self._scroll.set_shadow_type(Gtk.ShadowType.IN)

        self._store = Gtk.TreeStore(GdkPixbuf.Pixbuf, str, str, str, object)    # icon, message, file, line, Issue object

        self._view = Gtk.TreeView(model=self._store)

        column = Gtk.TreeViewColumn(_("Job"))

        pixbuf_renderer = Gtk.CellRendererPixbuf()
        column.pack_start(pixbuf_renderer, False)
        column.add_attribute(pixbuf_renderer, "pixbuf", 0)

        text_renderer = Gtk.CellRendererText()
        column.pack_start(text_renderer, True)
        column.add_attribute(text_renderer, "markup", 1)

        self._view.append_column(column)
        self._view.append_column(Gtk.TreeViewColumn(_("File"), Gtk.CellRendererText(), text=2))
        self._view.append_column(Gtk.TreeViewColumn(_("Line"), Gtk.CellRendererText(), text=3))

        self._handlers[self._view] = self._view.connect("row-activated", self._on_row_activated)

        self._scroll.add(self._view)
        self._scroll.set_hexpand(True)

        grid.add(self._scroll)

        # toolbar

        self._button_cancel = Gtk.ToolButton(stock_id=Gtk.STOCK_STOP)
        self._button_cancel.set_sensitive(False)
        self._button_cancel.set_tooltip_text(_("Abort Job"))
        self._handlers[self._button_cancel] = self._button_cancel.connect("clicked", self._on_abort_clicked)

        self._button_details = Gtk.ToolButton(stock_id=Gtk.STOCK_INFO)
        self._button_details.set_sensitive(False)
        self._button_details.set_tooltip_text(_("Show Detailed Output"))
        self._handlers[self._button_details] = self._button_details.connect("clicked", self._on_details_clicked)

        self._toolbar = Gtk.Toolbar()
        self._toolbar.set_style(Gtk.ToolbarStyle.ICONS)
        self._toolbar.set_icon_size(Gtk.IconSize.SMALL_TOOLBAR)
        self._toolbar.set_orientation(Gtk.Orientation.VERTICAL)
        self._toolbar.insert(self._button_cancel, -1)
        self._toolbar.insert(self._button_details, -1)
        self._toolbar.set_vexpand(True)

        grid.add(self._toolbar)

        # theme like gtk3
        ctx = self._scroll.get_style_context()
        ctx.set_junction_sides(Gtk.JunctionSides.RIGHT)

        ctx = self._toolbar.get_style_context()
        ctx.set_junction_sides(Gtk.JunctionSides.LEFT | Gtk.JunctionSides.RIGHT)
        ctx.add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)

        self.show_all()

    def get_label(self):
        return _("Tools")

    def get_icon(self):
        return Gtk.Image.new_from_stock(Gtk.STOCK_CONVERT, Gtk.IconSize.MENU)

    def _on_abort_clicked(self, button):
        self._abort_method.__call__()

    def _on_details_clicked(self, button):
        """
        The details button has been clicked
        """

    def _on_row_activated(self, view, path, column):
        issue = self._store.get(self._store.get_iter(path), 4)[0]
        if issue:
            self._context.activate_editor(issue.file)
            if self._context.active_editor:
                self._context.active_editor.select_lines(issue.start)
            else:
                LOG.error("No Editor object for calling select_lines")

    def clear(self):
        self._store.clear()

    def set_abort_enabled(self, enabled, method):
        # see issues.IStructuredIssueHandler.set_abort_enabled

        self._abort_method = method
        self._button_cancel.set_sensitive(enabled)

    def add_partition(self, label, state, parent_partition_id=None):
        """
        Add a new partition

        @param label: a label used in the UI
        @return: a unique id for the partition (here a Gtk.TreeIter)
        """
        icon = None
        if state == "running":
            icon = self._ICON_RUN
        elif state == "succeeded":
            icon = self._ICON_SUCCESS
        elif state == "failed":
            icon = self._ICON_FAIL
        elif state == "aborted":
            icon = self._ICON_ABORT

        self._view.expand_all()

        return self._store.append(parent_partition_id, [icon, label, "", "", None])

    def set_partition_state(self, partition_id, state):
        # see IStructuredIssueHandler.set_partition_state

        icon = None
        if state == "running":
            icon = self._ICON_RUN
        elif state == "succeeded":
            icon = self._ICON_SUCCESS
        elif state == "failed":
            icon = self._ICON_FAIL
        elif state == "aborted":
            icon = self._ICON_ABORT

        self._store.set_value(partition_id, 0, icon)

    def append_issues(self, partition_id, issues):
        for issue in issues:
            icon = None
            if issue.severity == Issue.SEVERITY_WARNING:
                icon = self._ICON_WARNING
            elif issue.severity == Issue.SEVERITY_ERROR:
                icon = self._ICON_ERROR
            self._store.append(partition_id, [icon, issue.message, issue.file.basename, str(issue.start), issue])

            LOG.debug("Issue: %s" % issue)

        self._view.expand_all()

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.completion
"""

import logging

from gi.repository import GdkPixbuf

from ..preferences import Preferences
from ..resources import Resources
from ..completion import ICompletionHandler, Proposal
from ..issues import MockIssueHandler
from .model import BibTeXModel
from .parser import BibTeXParser

LOG = logging.getLogger(__name__)

class BibTeXEntryTypeProposal(Proposal):
    """
    """

    _color = Preferences().get("light-foreground-color")

    def __init__(self, overlap, type):
        """
        @param overlap: the number of overlapping characters
        @param type: an EntryType
        """
        self._overlap = overlap
        self._type = type
        self._details = None
        self._source = None
        self._icon = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("document.png"))

    def _generate(self):
        """
        Generate snippet and details string
        """
        snippet = "@%s{${Identifier}" % self._type.name
        self._details = "@%s{<span color='%s'>Identifier</span>" % (self._type.name, self._color)
        for field in self._type.required_fields:
            snippet += ",\n\t%s = {${%s}}" % (field.name, field.label)
            self._details += ",\n\t%s = {<span color='%s'>%s</span>}" % (field.name, self._color, field.label)
        snippet += "\n}"
        self._details += "\n}"
        self._source = snippet

    @property
    def source(self):
        if not self._source:
            self._generate()
        return self._source

    @property
    def label(self):
        return self._type.name

    @property
    def details(self):
        if not self._details:
            self._generate()
        return self._details

    @property
    def icon(self):
        return self._icon

    @property
    def overlap(self):
        return self._overlap


class BibTeXCompletionHandler(ICompletionHandler):
    """
    This implements the BibTeX-specific code completion
    """
    trigger_keys = ["@"]
    prefix_delimiters = ["@"]

    def __init__(self):
        self._model = BibTeXModel()
        self._parser = BibTeXParser()
        self._issue_handler = MockIssueHandler()

    def complete(self, prefix):
        LOG.debug("BibTex complete: '%s'" % prefix)

        proposals = []

        if len(prefix) == 1:
            # propose all entry types
            types = self._model.types
            proposals = [BibTeXEntryTypeProposal(1, type) for type in types]
        else:
            if prefix[1:].isalpha():
                type_name_prefix = prefix[1:].lower()
                overlap = len(type_name_prefix) + 1
                # prefix is @[a-zA-Z]+
                types = [type for type in self._model.types if type.name.lower().startswith(type_name_prefix)]
                proposals = [BibTeXEntryTypeProposal(overlap, type) for type in types]
            else:
                # parse prefix
#                document = self._parser.parse(prefix, None, self._issue_handler)
#
#                print document
                pass

        return proposals



# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
This is searched by gedit for a class extending gedit.Plugin
"""

import logging
import string
from traceback import print_exc

from gi.repository import Gedit, GObject, Gio, Gtk, PeasGtk

from .preferences import Preferences
from .preferences.dialog import PreferencesDialog
from .preferences.tools import ToolPreferences
from .tools import ToolAction
from .resources import Resources
from .panelview import PanelView
from .config import EDITOR_VIEWS, ACTIONS
from .tabdecorator import GeditTabDecorator
from .windowcontext import WindowContext

LOG = logging.getLogger(__name__)

class LaTeXWindowActivatable(GObject.Object, Gedit.WindowActivatable, PeasGtk.Configurable):
    __gtype_name__ = "LaTeXWindowActivatable"

    """
    This class
     - manages the GeditTabDecorators
     - hooks the actions as menu items and tool items
     - installs side and bottom panel views
    """

    window = GObject.property(type=Gedit.Window)

    # ui definition template for hooking tools in Gedit's ui.
    _tool_ui_template = string.Template("""<ui>
            <menubar name="MenuBar">
                <menu name="ToolsMenu" action="ToolsDummyAction">
                    <placeholder name="ToolsOps_1">$items</placeholder>
                </menu>
            </menubar>
            <toolbar name="$toolbar_name">
                <toolitem action="LaTeXBuildAction">
                    <menu action="LaTeXBuildMenuAction">
                        <placeholder name="LaTeXBuildPlaceholder_1">$items</placeholder>
                    </menu>
                </toolitem>
            </toolbar>
        </ui>""")

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        """
        Called when the window extension is activated
        """

        self._preferences = Preferences()
        self._tool_preferences = ToolPreferences()
        self._tool_preferences.connect("tools-changed", self._on_tools_changed)

        #
        # initialize context object
        #
        self._window_context = WindowContext(self, EDITOR_VIEWS)

        # the order is important!
        self._init_actions()
        self._init_tool_actions()
        self._init_views()
        self._init_tab_decorators()

        #
        # listen to tab signals
        #
        self._signal_handlers = [
                self.window.connect("tab_added", self._on_tab_added),
                self.window.connect("tab_removed", self._on_tab_removed),
                ]

    def do_deactivate(self):
        """
        Called when the window extension is deactivated
        """
        # save preferences and stop listening
        self._tool_preferences.save()

        # destroy tab decorators
        self._active_tab_decorator = None
        for tab in self._tab_decorators:
            self._tab_decorators[tab].destroy()
        self._tab_decorators = {}

        # disconnect from tab signals
        for handler in self._signal_handlers:
            self.window.disconnect(handler)
        del self._signal_handlers

        # remove all views
        self.disable()

        # remove toolbar
        if self._toolbar:
            self._toolbar.destroy()

        # remove tool actions
        self._ui_manager.remove_ui(self._tool_ui_id)
        for gtk_action in self._action_handlers:
            gtk_action.disconnect(self._action_handlers[gtk_action])
            self._tool_action_group.remove_action(gtk_action)
        self._ui_manager.remove_action_group(self._tool_action_group)

        # remove actions
        self._ui_manager.remove_ui(self._ui_id)
        for clazz in self._action_objects:
            self._action_objects[clazz].unhook(self._action_group)
        self._ui_manager.remove_action_group(self._action_group)

        # destroy the window context
        self._window_context.destroy()
        del self._window_context

    def do_create_configure_widget(self):
        return PreferencesDialog().dialog

    def _init_views(self):
        """
        """

        # selection states for each TabDecorator
        self._selected_bottom_views = {}
        self._selected_side_views = {}

        # currently hooked editor views
        self._side_views = []
        self._bottom_views = []

    def _init_actions(self):
        """
        Merge the plugin's UI definition with the one of Gedit and hook the actions
        """
        self._ui_manager = Gtk.UIManager()
        self._action_group = Gtk.ActionGroup("LaTeXWindowActivatableActions")
        self._icon_factory = Gtk.IconFactory()
        self._icon_factory.add_default()

        # create action instances, hook them and build up some
        # hash tables

        self._action_objects = {}        # name -> Action object
        self._action_extensions = {}    # extension -> action names

        for clazz in ACTIONS:
            action = clazz(icon_factory=self._icon_factory)
            action.hook(self._action_group, self._window_context)
            action.simplehook(self.window, self._window_context)

            self._action_objects[clazz.__name__] = action

            for extension in action.extensions:
                if extension in list(self._action_extensions.keys()):
                    self._action_extensions[extension].append(clazz.__name__)
                else:
                    self._action_extensions[extension] = [clazz.__name__]

        #toolbar_mode = self._preferences.get("toolbar-mode")
        # force normal mode
        toolbar_mode = "normal"
        
        # merge ui
        self._ui_manager.insert_action_group(self._action_group, -1)
        self._ui_id = self._ui_manager.add_ui_from_file(Resources().get_ui_file("ui-toolbar-%s.builder" % toolbar_mode))

        if toolbar_mode == "normal":
            self._toolbar = self._ui_manager.get_widget("/LaTeXToolbar")
            self._toolbar_name = "LaTeXToolbar"
            self._toolbar.set_style(Gtk.ToolbarStyle.BOTH_HORIZ)
            # FIXME: Adding a new toolbar to gedit is not really public API
            try:
                # We assume the spot below exists
                self._main_box = self.window.get_children()[0].get_children()[0].get_children()[0].get_children()[1]
                self._main_box.pack_start(self._toolbar, False, True, 0)
                self._main_box.reorder_child(self._toolbar, 1)
            except (IndexError, AttributeError, TypeError) as e:
                print_exc()
                print()
                print("Could not find place for the toolbar. Disabling toolbar.")
                self_toolbar = None
                self._toolbar_name = ""
        elif toolbar_mode == "combined":
            self._toolbar = None
            self._toolbar_name = "ToolBar"
        else:
            self._toolbar = None
            self._toolbar_name = ""
            LOG.info("Toolbar disabled")    
            
    def _init_tab_decorators(self):
        """
        Look for already open tabs and create decorators for them
        """
        self._tab_decorators = {}
        self._active_tab_decorator = None
        active_view = self.window.get_active_view()
        views = self.window.get_views()

        for view in views:
            tab = Gedit.Tab.get_from_document(view.get_buffer())
            decorator = self._create_tab_decorator(tab, init=True)
            if view is active_view:
                self._active_tab_decorator = decorator

        LOG.debug("_init_tab_decorators: initialized %s decorators" % len(views))

        if len(views) > 0 and not self._active_tab_decorator:
            LOG.warning("_init_tab_decorators: no active decorator found")

    def _init_tool_actions(self):
        """
         - Load defined Tools
         - create and init ToolActions from them
         - hook them in the window UI
         - create a map from extensions to lists of ToolActions
        """
        items_ui = ""
        self._action_handlers = {}

        # this is used for enable/disable actions by name (None = every extension)
        self._tool_action_extensions = {None: []}
        self._tool_action_group = Gtk.ActionGroup("LaTeXPluginToolActions")

        i = 1                    # counting tool actions
        accel_counter = 1        # counting tool actions without custom accel
        for tool in self._tool_preferences.tools:
            # hopefully unique action name
            name = "Tool%sAction" % i

            # update extension-tool mapping
            for extension in tool.extensions:
                try:
                    self._tool_action_extensions[extension].append(name)
                except KeyError:
                    # extension not yet mapped
                    self._tool_action_extensions[extension] = [name]

            # create action
            action = ToolAction(tool)
            gtk_action = Gtk.Action(name, action.label, action.tooltip, action.stock_id)
            self._action_handlers[gtk_action] = gtk_action.connect("activate", lambda gtk_action, action: action.activate(self._window_context), action)
            
            # create simple actions to be used by menu (created in appactivatable.py)
            simpleaction = Gio.SimpleAction(name=name)
            simpleaction.connect("activate", lambda _a, _b, action: action.activate(self._window_context), action)
            self.window.add_action(simpleaction)

            accelerator = None
            if tool.accelerator and len(tool.accelerator) > 0:
                key,mods = Gtk.accelerator_parse(tool.accelerator)
                if Gtk.accelerator_valid(key,mods):
                    accelerator = tool.accelerator
            if not accelerator:
                accelerator = "<Ctrl><Alt>%s" % accel_counter
                accel_counter += 1
            self._tool_action_group.add_action_with_accel(gtk_action, accelerator)

            # add UI definition
            items_ui += """<menuitem action="%s" />""" % name
            i += 1
        items_ui +="""<separator/>"""

        tool_ui = self._tool_ui_template.substitute({
                            "items": items_ui,
                            "toolbar_name": self._toolbar_name})
        
        self._ui_manager.insert_action_group(self._tool_action_group, -1)
        self._tool_ui_id = self._ui_manager.add_ui_from_string(tool_ui)
        
    def save_file(self):
        """
        Trigger the 'Save' action

        (used by ToolAction before tool run)
        """
        tab = self._active_tab_decorator.tab
        Gedit.commands_save_document(tab.get_toplevel(), tab.get_document())

    def show_toolbar(self):
        if self._toolbar:
            self._toolbar.show()

    def hide_toolbar(self):
        if self._toolbar:
            self._toolbar.hide()

    def _on_tools_changed(self, tools):
        LOG.debug("_on_tools_changed")

        # remove tool actions and ui
        self._ui_manager.remove_ui(self._tool_ui_id)
        for gtk_action in self._action_handlers:
            gtk_action.disconnect(self._action_handlers[gtk_action])
            self._tool_action_group.remove_action(gtk_action)
        self._ui_manager.remove_action_group(self._tool_action_group)

        # re-init tool actions
        self._init_tool_actions()

        # re-adjust action states
        GObject.idle_add(self.adjust,self._active_tab_decorator)

    def activate_tab(self, file):
        """
        Activate the GeditTab containing the given File or open a new
        tab for it (this is called by the WindowContext)

        @param file: a File object
        """
        for tab, tab_decorator in self._tab_decorators.items():
            if tab_decorator.file and tab_decorator.file == file:
                self.window.set_active_tab(tab)
                return

        # not found, open file in a new tab...

        uri = file.uri
        gfile = Gio.file_new_for_uri(uri)

        if Gedit.utils_is_valid_location(gfile):
            LOG.debug("GeditWindow.create_tab_from_uri(%s)" % uri)
            try:
                view = self.window.get_active_view()
                encoding = view.get_buffer().get_encoding()
            except AttributeError:
                # No document open
                encoding = None
            self.window.create_tab_from_location(
                            gfile, encoding,
                            1, 1, False, True)
        else:
            LOG.error("Gedit.utils.uri_is_valid(%s) = False" % uri)

    def disable(self):
        """
        Called if there are no more tabs after tab_removed
        """
        if self._toolbar:
            self._toolbar.hide()

        # disable all actions
        for name in self._action_objects.keys():
            self._action_group.get_action(name).set_visible(False)

        # disable all tool actions
        for l in list(self._tool_action_extensions.values()):
            for name in l:
                self._tool_action_group.get_action(name).set_sensitive(False)

    def adjust(self, tab_decorator):
        """
        Adjust actions and views according to the currently active TabDecorator
        (the file type it contains)

        Called by
         * do_update_state()
         * GeditTabDecorator when the Editor instance changes
        """

        # TODO: improve and simplify this!
        extension = tab_decorator.extension

        LOG.debug("---------- ADJUST: %s" % (extension))

        latex_extensions = self._preferences.get("latex-extensions").split(",")

        #
        # adjust actions
        #
        # FIXME: we always get the state of the new decorator after tab change
        # but we need to save the one of the old decorator
        #
        # FIXME: we are dealing with sets so saving the index as selection state
        # is nonsense
        #
        
        for category, catalog in (('actions', self._action_extensions),
                                  ('tools', self._tool_action_extensions)):
            # enable actions for all extensions
            to_enable = set(catalog[None])
            # enable the actions registered for the extension
            if extension:
                try:
                    to_enable.update(catalog[extension])
                except KeyError:
                    pass
            for name in set().union(*catalog.values()):
                action = self.window.lookup_action(name)
                action.set_enabled(name in to_enable)
            
            if category == 'tools':
                # Show/hide the tool menu itself onlyif any action is shown:
                menu_action = self.window.lookup_action('ToolsDummyAction')
                # FIXME: This line spits "Gtk-WARNING"s which I really don't
                # understand (when "to_enable" is not empty):
                # "Duplicate child name in GtkStack: LaTeX Tools"
                menu_action.set_enabled(bool(to_enable))
        
                if to_enable:
                    self.show_toolbar()
                else:
                    self.hide_toolbar()

        #
        # adjust editor views
        #

        side_views = []
        bottom_views = []

        if tab_decorator.editor:
            editor_views = self._window_context.editor_views[tab_decorator.editor]
            for id, view in editor_views.items():
                if isinstance(view, PanelView):
                    if view.get_orientation() == Gtk.Orientation.HORIZONTAL:
                        bottom_views.append(view)
                    else:
                        side_views.append(view)
                else:
                    raise RuntimeError("Invalid view type: %s" % view)

        for view in self._side_views:
            view.hide()

        for view in self._bottom_views:
            view.hide()

        # show all current views
        i = len(self._side_views)
        for view in side_views:
            if view in self._side_views:
                view.show()
            else:
                self.window.get_side_panel().add_titled(view, "after_side_view_id" + str(i), view.get_label())
                self._side_views.append(view)
                i += 1

        i = len(self._bottom_views)
        for view in bottom_views:
            if view in self._bottom_views:
                view.show()
            else:
                self.window.get_bottom_panel().add_titled(view, "bottom_view_id" + str(i), view.get_label())
                self._bottom_views.append(view)
                i += 1

    def _on_tab_added(self, window, tab):
        """
        A new tab has been added

        @param window: Gedit.Window object
        @param tab: Gedit.Tab object
        """
        LOG.debug("tab_added")

        if tab in self._tab_decorators:
            LOG.warning("There is already a decorator for tab %s" % tab)
            return

        self._create_tab_decorator(tab)

    def _on_tab_removed(self, window, tab):
        """
        A tab has been closed

        @param window: GeditWindow
        @param tab: the closed GeditTab
        """
        LOG.debug("tab_removed")

        # As we don't call GeditWindowDecorator.adjust() if the new
        # tab is not the active one (for example, when opening several
        # files at once, see GeditTabDecorator._adjust_editor()),
        # it may happen that self._selected_side_views[tab] is not set.
        if self._tab_decorators[tab] in self._selected_side_views:
            del self._selected_side_views[self._tab_decorators[tab]]
        if self._tab_decorators[tab] in self._selected_bottom_views:
            del self._selected_bottom_views[self._tab_decorators[tab]]

        self._tab_decorators[tab].destroy()
        if self._active_tab_decorator == self._tab_decorators[tab]:
            self._active_tab_decorator = None

        del self._tab_decorators[tab]

        if len(self._tab_decorators) == 0:
            # no more tabs
            self.disable()

    def do_update_state(self):
        """
        The active tab has changed

        """
        
        tab = self.window.get_active_tab()
        if not tab:
            # Initialization time
            return
        LOG.debug("active_tab_changed")

        if tab in list(self._tab_decorators.keys()):
            decorator = self._tab_decorators[tab]
        else:
            # (on Gedit startup 'tab-changed' comes before 'tab-added')
            # remember: init=True crashes the plugin here!
            decorator = self._create_tab_decorator(tab)

        self._active_tab_decorator = decorator

        # adjust actions and views
        GObject.idle_add(self.adjust, decorator)

    def _create_tab_decorator(self, tab, init=False):
        """
        Create a new GeditTabDecorator for a GeditTab
        """
        decorator = GeditTabDecorator(self, tab, init)
        self._tab_decorators[tab] = decorator
        return decorator

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
outline

Classes used for creating an outline view of LaTeX and BibTeX files
"""

import logging

from gi.repository import Gtk, GdkPixbuf

from .panelview import PanelView
from .preferences import Preferences
from .resources import Resources
from .gldefs import _

LOG = logging.getLogger(__name__)

class BaseOutlineView(PanelView):
    """
    Base class for the BibTeX and LaTeX outline views
    """

    def __init__(self, context, editor):
        PanelView.__init__(self, context)
        self._editor = editor

        self.set_orientation(Gtk.Orientation.VERTICAL)

        self._preferences = Preferences()

        grid = Gtk.Grid()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.add(grid)

        # toolbar

        btn_follow = Gtk.ToggleToolButton.new_from_stock(Gtk.STOCK_CONNECT)
        btn_follow.set_tooltip_text(_("Follow Editor"))
        btn_follow.set_active(self._preferences.get("outline-connect-to-editor"))
        btn_follow.connect("toggled", self._on_follow_toggled)

        btn_expand = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_IN)
        btn_expand.set_tooltip_text(_("Expand All"))
        btn_expand.connect("clicked", self._on_expand_clicked)

        btn_collapse = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_OUT)
        btn_collapse.set_tooltip_text(_("Collapse All"))
        btn_collapse.connect("clicked", self._on_collapse_clicked)

        self._toolbar = Gtk.Toolbar()
        self._toolbar.set_style(Gtk.ToolbarStyle.ICONS)
        self._toolbar.set_icon_size(Gtk.IconSize.MENU)
        self._toolbar.insert(btn_follow, -1)
        self._toolbar.insert(Gtk.SeparatorToolItem(), -1)
        self._toolbar.insert(btn_expand, -1)
        self._toolbar.insert(btn_collapse, -1)
        self._toolbar.insert(Gtk.SeparatorToolItem(), -1)
        self._toolbar.set_hexpand(True)

        grid.add(self._toolbar)

        # tree view

        column = Gtk.TreeViewColumn()

        pixbuf_renderer = Gtk.CellRendererPixbuf()
        column.pack_start(pixbuf_renderer, False)
        column.add_attribute(pixbuf_renderer, "pixbuf", 1)

        text_renderer = Gtk.CellRendererText()
        column.pack_start(text_renderer, True)
        column.add_attribute(text_renderer, "markup", 0)

        self._offset_map = OutlineOffsetMap()

        self._store = Gtk.TreeStore(str, GdkPixbuf.Pixbuf, object)    # label, icon, node object

        self._view = Gtk.TreeView(model=self._store)
        self._view.append_column(column)
        self._view.set_headers_visible(False)
        self._view.connect("row-activated", self._on_row_activated)

        scrolled = Gtk.ScrolledWindow()
        scrolled.add(self._view)
        scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled.set_vexpand(True)

        grid.add(scrolled)

        # this holds a list of the currently expanded paths
        self._expandedPaths = None

    def get_label(self):
        return _("Outline")

    def get_icon(self):
        return Gtk.Image.new_from_file(Resources().get_icon("outline.png"))

    def _on_follow_toggled(self, toggle_button):
        value = toggle_button.get_active()
        self._preferences.set("outline-connect-to-editor", value)

    def _on_expand_clicked(self, button):
        self._view.expand_all()

    def _on_collapse_clicked(self, button):
        self._view.collapse_all()

    def select_path_by_offset(self, offset):
        """
        Select the path corresponding to a given offset in the source

        Called by the Editor
        """
        LOG.debug("select path by offset %r" % offset)
        try:
            path = self._offset_map.lookup(offset)
            selection = self._view.get_selection()
            if selection:
                self._view.expand_to_path(path)
                selection.select_path(path)
        except KeyError:
            pass

    def _save_state(self):
        """
        Save the current expand state
        """
        self._expanded_paths = []
        self._view.map_expanded_rows(self._save_state_map_function,None)

    def _save_state_map_function(self, view, path, user_data=None):
        """
        Mapping function for saving the current expand state
        """
        self._expanded_paths.append(path.to_string())

    def _restore_state(self):
        """
        Restore the last expand state
        """
        self._view.collapse_all()

        if self._expanded_paths:
            for path in self._expanded_paths:
                self._view.expand_to_path(Gtk.TreePath.new_from_string(path))

    def _on_row_activated(self, view, path, column):
        it = self._store.get_iter(path)
        node = self._store.get(it, 2)[0]

        self._on_node_activated(node)

    #
    # methods to be overridden by the subclass
    #

    def _on_node_activated(self, node):
        """
        To be overridden
        """

    def set_outline(self, outline):
        """
        Load a new outline model

        To be overridden
        """


class Item(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __lt__(self, other):
        return self.key < other.key

    def __eq__(self, other):
        return self.key == other.key

    def __hash__(self):
        return hash(self.key)


class OffsetLookupTree(object):
    def __init__(self):
        self.items = []

    def insert(self, key, value):
        """
        Insert a value
        """
        self.items.append(Item(key, value))

    def prepare(self):
        """
        Prepare the structure for being searched
        """
        self.items.sort()

    def find(self, key):
        """
        Find a value by its key
        """
        return self._find(key, 0, len(self.items) - 1)

    def _find(self, key, lo, hi):
        if hi - lo == 0:
            raise KeyError

        i = (hi - lo)/2
        item = self.items[i]
        if item.key == key:
            return item.value
        elif item.key > key:
            return self._find(key, lo, i - 1)
        elif item.key < key:
            return self._find(key, i + 1, hi)


class OutlineOffsetMap(object):
    """
    This stores a mapping from the start offsets of outline elements
    to paths in the outline tree.

    We need this for the 'connect-outline-to-editor' feature.

    We may not use a simple dictionary for that because we need some
    kind of neighborhood lookup as we never get the exact offsets.
    """

    # TODO: find a faster structure for this - this is a tree with a special
    # find method

    def __init__(self):
        self._map = {}

    def clear(self):
        self._map = {}

    def put(self, offset, path):
        self._map[offset] = path

    def lookup(self, offset):
        """
        Returns the outline element containing a certain offset
        """

        # sort offsets
        offsets = list(self._map.keys())
        offsets.sort()

        # find nearest offset
        nearestOffset = None
        for o in offsets:
            if o > offset:
                break
            nearestOffset = o

        if not nearestOffset:
            raise KeyError

        return self._map[nearestOffset]

    def __str__(self):
        s = "<OutlineOffsetMap>"

        ofs = list(self._map.keys())
        ofs.sort()

        for o in ofs:
            s += "\n\t%s : %s" % (o, self._map[o])
        s += "\n</OutlineOffsetMap>"
        return s

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.cache

Cache for LaTeX document models for speeding up reference expanding
"""

from logging import getLogger

from .parser import LaTeXParser
from ..issues import IIssueHandler


class CacheIssueHandler(IIssueHandler):
    """
    This is used to catch the issues occuring while parsing a document
    on cache fault.
    """
    def __init__(self):
        self.issues = []

    def clear(self):
        self.issues = []

    def issue(self, issue):
        self.issues.append(issue)


class LaTeXDocumentCache(object):
    """
    This caches LaTeX document models. It used to speed up
    the LaTeXReferenceExpander.
    """

    # FIXME: we need a global character set

    # TODO: serialize the cache on shutdown

    _log = getLogger("LaTeXDocumentCache")

    class Entry(object):
        """
        An entry in the cache
        """
        _log = getLogger("LaTeXDocumentCache.Entry")

        def __init__(self, file, charset):
            self.__file = file
            self.__parser = LaTeXParser()
            self.__issue_handler = CacheIssueHandler()
            self.__mtime = 0
            self.__document = None
            self.__charset = charset

            self.synchronize()

        @property
        def modified(self):
            return (self.__file.mtime > self.__mtime)

        @property
        def document(self):
            return self.__document

        @property
        def issues(self):
            return self.__issue_handler.issues

        def synchronize(self):
            """
            Synchronize document model with file contents.

            @raise OSError: if the file is not found
            """
            # update timestamp
            self.__mtime = self.__file.mtime

            # clear previous data
            self.__issue_handler.clear()
            if self.__document != None:
                self.__document.destroy()
                self.__document = None

            # read file
            try:
                f = open(self.__file.path, "r", encoding=self.__charset)
                try:
                    content = f.read()
                finally:
                    f.close()
            except IOError:
                return

            # parse
            self.__document = self.__parser.parse(content, self.__file, self.__issue_handler)

    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):
            self._entries = {}
            self._ready = True

    def get_document(self, file, charset, issue_handler):
        """
        Return the (hopefully) cached document model for a given file

        @param file: a File object
        @param charset: character set
        @param issue_handler: an IIssueHandler to use
        """
        try:
            # update entry if necessary
            entry = self._entries[file.uri]
            self._log.debug("Reading '%s' from cache" % file)
            if entry.modified:
                self._log.debug("File '%s' modified, synchronizing..." % file)
                entry.synchronize()
        except KeyError:
            self._log.debug("Cache fault for '%s'" % file)
            # create new entry
            entry = self.Entry(file, charset)
            self._entries[file.uri] = entry

        # pass cached issues to the issue handler
        for issue in entry.issues:
            issue_handler.issue(issue)

        return entry.document



# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.completion

LaTeX-specific completion classes
"""

from logging import getLogger
from gi.repository import GdkPixbuf

from ..resources import Resources
from ..completion import ICompletionHandler, Proposal


class LaTeXCommandProposal(Proposal):
    """
    A proposal inserting a Template when activated
    """

    def __init__(self, overlap, snippet, label):
        self._snippet = snippet
        self._label = label
        self._overlap = overlap
        self._icon = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("i_command.png"))

    @property
    def source(self):
        return self._snippet

    @property
    def label(self):
        return self._label

    @property
    def details(self):
        return None

    @property
    def icon(self):
        return self._icon

    @property
    def overlap(self):
        return self._overlap


class LaTeXChoiceProposal(Proposal):
    """
    A proposal inserting a simple string when activated
    """

    def __init__(self, overlap, source, label, details):
        self._source = source
        self._details = details
        self._overlap = overlap
        self._label = label
        self._icon = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("i_choice.png"))

    @property
    def source(self):
        return self._source

    @property
    def label(self):
        return self._label

    @property
    def details(self):
        return self._details

    @property
    def icon(self):
        return self._icon

    @property
    def overlap(self):
        return self._overlap


from .model import LanguageModelFactory, Choice, MandatoryArgument, OptionalArgument
from .parser import PrefixParser, Node

from ..bibtex.cache import BibTeXDocumentCache


class LaTeXCompletionHandler(ICompletionHandler):
    """
    This implements the LaTeX-specific code completion
    """
    _log = getLogger("LaTeXCompletionHandler")

    trigger_keys = ["backslash", "braceleft"]
    prefix_delimiters = ["\\"]

    def __init__(self):
        self._log.debug("init")
        #get the language_model singleton
        self._language_model = LanguageModelFactory().get_language_model()
        self._bibtex_document_cache = BibTeXDocumentCache()

    def set_outline(self, outline):
        """
        Process a LaTeX outline model

        @param outline: a latex.outline.Outline instance
        """

        # labels
        label_choices = [Choice(None, label.value) for label in outline.labels]
        self._language_model.fill_placeholder("Labels", label_choices)

        # colors
        color_choices = [Choice(None, color) for color in outline.colors]
        self._language_model.fill_placeholder("Colors", color_choices)

        # newcommands
        self._language_model.set_newcommands(outline.newcommands)

        # newenvironments
        newenvironments = []
        for n in outline.newenvironments:
            choice = Choice(None, n.value)
            newenvironments.append(choice)
        self._language_model.fill_placeholder("Newenvironments", newenvironments)

        #
        # bibtex entries
        #
        try:

            entry_choices = []

            for bib_file in outline.bibliographies:
                try:
                    bibtex_document = self._bibtex_document_cache.get_document(bib_file)

                    # generate choices from entries

                    for entry in bibtex_document.entries:

                        # build table data for DetailsPopup
                        rows = []
                        for field in entry.fields:
                            rows.append([field.name, field.valueMarkup])

                        entry_choices.append(Choice(None, entry.key, rows))

                except OSError:
                    # BibTeX file not found
                    self._log.error("Not found: %s" % bib_file)

            # attach to placeholders in CommandStore
            self._language_model.fill_placeholder("Bibitems", entry_choices)

        except IOError:
            self._log.debug("Failed to provide BibTeX completion due to IOError")


    def set_neighbors(self, tex_files, bib_files, graphic_files):
        """
        Populate the lists of neighbor files

        @param tex_files: list of neighbor TeX files
        @param bib_files: list of neighbor BibTeX files
        @param graphic_files: list of neighbor graphics
        """
        tex_choices = [Choice(None, file.shortbasename) for file in tex_files]
        self._language_model.fill_placeholder("TexFiles", tex_choices)

        bib_choices = [Choice(None, file.shortbasename) for file in bib_files]
        self._language_model.fill_placeholder("BibFiles", bib_choices)

        graphic_choices = [Choice(None, file.basename) for file in graphic_files]
        self._language_model.fill_placeholder("ImageFiles", graphic_choices)

    def complete(self, prefix):
        """
        Try to complete a given prefix
        """
        self._log.debug("complete: '%s'" % prefix)

        #proposals = [LaTeXTemplateProposal(Template("Hello[${One}][${Two}][${Three}]"), "Hello[Some]"), LaTeXProposal("\\world")]

        fragment = Node(Node.DOCUMENT)
        parser = PrefixParser()

        try:
            parser.parse(prefix, fragment)

            modelParser = PrefixModelParser(self._language_model)
            proposals = modelParser.parse(fragment)

            self._log.debug("Generated %s proposals" % len(proposals))

            return proposals

        except Exception as e:
            self._log.debug(e)

        return []


from ..preferences import Preferences
from . import LaTeXSource


class PrefixModelParser(object):
    """
    This parses the document model of a prefix and generates proposals accordingly

    This is used by the LaTeXCompletionHandler class
    """

    _log = getLogger("PrefixModelParser")

    def __init__(self, language_model):
        self.__language_model = language_model
        self.__light_foreground = Preferences().get("light-foreground-color")

    def __create_proposals_from_commands(self, commands, overlap):
        """
        Generate proposals for commands
        """
        proposals = []

        for command in commands:
            label = command.name
            snippet = "\\" + command.name

            for idx, argument in enumerate(command.children):
                if type(argument) is MandatoryArgument:
                    label += "{<span color='%s'>%s</span>}" % (self.__light_foreground, argument.label)
                    snippet += "{${%s:%s}}" % (idx+1, argument.label)
                elif type(argument) is OptionalArgument:
                    label += "[<span color='%s'>%s</span>]" % (self.__light_foreground, argument.label)
                    snippet += "[${%s:%s}]" % (idx+1, argument.label)

            if command.package:
                label += " <small><b>%s</b></small>" % command.package

            # workaround for latex.model.Element.package may be None
            # TODO: latex.model.Element.package should be a list of packages
            if command.package is None:
                packages = []
            else:
                packages = [command.package]
            proposal = LaTeXCommandProposal(overlap, LaTeXSource(snippet, packages), label)
            proposals.append(proposal)

        return proposals

    def __create_proposals_from_choices(self, choices, overlap):
        """
        Generate proposals for argument choices
        """
        proposals = []

        for choice in choices:
            label = choice.value
            if choice.package:
                label += " <small><b>%s</b></small>" % choice.package

            # see above
            if choice.package is None:
                packages = []
            else:
                packages = [choice.package]
            proposal = LaTeXChoiceProposal(overlap, LaTeXSource(choice.value, packages), label, choice.details)
            proposals.append(proposal)

        return proposals

    def parse(self, prefixFragment):
        """
        Returns choices
        """

        # root node of the prefix model must be COMMAND
        commandNode = prefixFragment[-1]
        if commandNode.type != Node.COMMAND:
            return []

        commandName = commandNode.value

        if len(commandNode) == 0:
            # command has no arguments...

            if len(commandName) == 0:
                # no name, so propose all commands
                commands = list(self.__language_model.commands.values())
                overlap = 1        # only "\"
            else:
                commands = self.__language_model.find_command(commandName)

                if len(commands) == 1 and commands[0].name == commandName:
                    # don't propose when only one command is found and that one
                    # matches the typed one
                    return []

                overlap = len(commandName) + 1         # "\begi"

            return self.__create_proposals_from_commands(commands, overlap)

        # ...command has arguments

        try:
            self._log.debug(commandNode.xml)

            # find the language model of the command
            storedCommand = self.__language_model.commands[commandName]


            try:
                argumentNode, storedArgument = self.__match_argument(commandNode, storedCommand)
            except Exception as e:
                self._log.error(e)
                return []


            choices = storedArgument.children

            # filter argument matching the already typed argument text

            argumentValue = argumentNode.innerText

            if len(argumentValue):
                choices = [choice for choice in choices if choice.value.startswith(argumentValue)]
                overlap = len(argumentValue)
            else:
                overlap = 0
            return self.__create_proposals_from_choices(choices, overlap)

        except KeyError:
            self._log.debug("Command not found: %s" % commandName)
            return []

    def __match_argument(self, command, model_command):
        """
        @param command: the parsed command Node
        @param model_command: the according model command
        @return: (matched argument, model argument)
        """
        # push the arguments of the model command on a stack
        model_argument_stack = []
        model_argument_stack.extend(model_command.children)
        model_argument_stack.reverse()

        for argument in command:
            if argument.type == Node.MANDATORY_ARGUMENT:

                # optional arguments in the model may be skipped
                while True:
                    try:
                        model_argument = model_argument_stack.pop()
                        if model_argument.type != Node.OPTIONAL_ARGUMENT:
                            break
                    except IndexError:
                        # no more optional arguments to skip - signatures can't match
                        raise Exception("Signatures don't match")

                if not argument.closed:
                    return (argument, model_argument)

            elif argument.type == Node.OPTIONAL_ARGUMENT:
                model_argument = model_argument_stack.pop()

                if model_argument.type != Node.OPTIONAL_ARGUMENT:
                    raise Exception("Signatures don't match")

                if not argument.closed:
                    return (argument, model_argument)

        raise Exception("No matching model argument found")


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.actions
"""
from gi.repository import Gtk

from logging import getLogger

from ..action import Action, IconAction
from ..preferences import Preferences
from ..issues import MockIssueHandler
from ..tools import ToolRunner
from ..gldefs import _
from .editor import LaTeXEditor
from .parser import LaTeXParser, Node
from .dialogs import UseBibliographyDialog, InsertGraphicsDialog, InsertTableDialog, \
                    InsertListingDialog, BuildImageDialog, \
                    NewDocumentDialog
from . import LaTeXSource

LOG = getLogger(__name__)

class LaTeXAction(Action):
    extensions = Preferences().get("latex-extensions").split(",")


class LaTeXIconAction(IconAction):
    extensions = Preferences().get("latex-extensions").split(",")


class LaTeXTemplateAction(LaTeXIconAction):
    """
    Utility base class for quickly defining Actions inserting a LaTeX snippet
    """
    accelerator = None
    snippet_source = None
    packages = []

    def activate(self, context):
        context.active_editor.insert(LaTeXSource(self.snippet_source, self.packages))


class LaTeXMenuAction(LaTeXAction):
    label = _("LaTeX")
    stock_id = None
    accelerator = None
    tooltip = None

    def activate(self, context):
        pass


class LaTeXNewAction(Action):
    label = _("New LaTeX Document...")
    stock_id = Gtk.STOCK_NEW
    accelerator = "<Ctrl><Alt>N"
    tooltip = _("Create a new LaTeX document")

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = NewDocumentDialog()

        # we may not open the empty file and insert a Temlate here
        # because WindowContext.activate_editor calls gedit.Window.create_tab_from_uri
        # which is async

        if self.dialog.run() == 1:
            file = self.dialog.file
            file.create(self.dialog.source)
            context.activate_editor(file)


class LaTeXChooseMasterAction(LaTeXAction):
    label = _("Choose Master Document...")
    stock_id = None
    accelerator = None
    tooltip = None

    def activate(self, context):
        editor = context.active_editor
        assert type(editor) is LaTeXEditor

        editor.choose_master_file()

class LaTeXCloseEnvironmentAction(LaTeXIconAction):
    label = _("Close Nearest Environment")
    accelerator = "<Ctrl><Alt>E"
    tooltip = _("Close the nearest TeX environment at left of the cursor")
    icon_name = "close_env"

    def activate(self, context):
        # FIXME: use the document model of the Editor

        editor = context.active_editor
        assert type(editor) is LaTeXEditor

        # push environments on stack and find nearest one to close
        try:
            self._stack = []
            self._find_open_environments(LaTeXParser().parse(editor.content_at_left_of_cursor, None, MockIssueHandler()))

            if len(self._stack) > 0:
                editor.insert("\\end{%s}" % self._stack[-1])
            else:
                LOG.debug("No environment to close")
        except ValueError:
            LOG.info("Environments are malformed")

    def _find_open_environments(self, parent_node):
        for node in parent_node:
            recurse = True
            if node.type == Node.COMMAND:
                if node.value == "begin":
                    # push environment on stack
                    environ = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                    self._stack.append(environ)

                elif node.value == "end":
                    # pop from stack
                    environ = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                    try:
                        top_environ = self._stack.pop()
                        if top_environ != environ:
                            raise ValueError()
                    except IndexError:
                        raise ValueError()

                elif node.value == "newcommand":
                    recurse = False

            if recurse:
                self._find_open_environments(node)


class LaTeXUseBibliographyAction(LaTeXIconAction):
    label = _("Use Bibliography...")
    accelerator = None
    tooltip = _("Use Bibliography")
    icon_name = "bib"

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = UseBibliographyDialog()

        source = self.dialog.run_dialog(context.active_editor.edited_file)
        if source:
            editor = context.active_editor
            assert type(editor) is LaTeXEditor

            editor.insert_at_position(source + "\n\n", LaTeXEditor.POSITION_BIBLIOGRAPHY)


class LaTeXFontFamilyAction(LaTeXIconAction):
    menu_tool_action = True

    label = _("Font Family")
    accelerator = None
    tooltip = _("Font Family")
    icon_name = "bf"

    def activate(self, context):
        pass


class LaTeXFontFamilyMenuAction(LaTeXMenuAction):
    label = _("Font Family")
    tooltip = _("Font Family")


class LaTeXBoldAction(LaTeXTemplateAction):
    label = _("Bold")
    tooltip = _("Bold")
    icon_name = "bf"
    snippet_source = "\\textbf{$0}"


class LaTeXItalicAction(LaTeXTemplateAction):
    label = _("Italic")
    tooltip = _("Italic")
    icon_name = "it"
    snippet_source = "\\textit{$0}"


class LaTeXEmphasizeAction(LaTeXTemplateAction):
    label = _("Emphasize")
    tooltip = _("Emphasize")
    icon_name = "it"
    snippet_source = "\\emph{$0}"


class LaTeXUnderlineAction(LaTeXTemplateAction):
    label = _("Underline")
    tooltip = _("Underline")
    icon_name = "underline"
    snippet_source = "\\underline{$0}"


class LaTeXSmallCapitalsAction(LaTeXTemplateAction):
    label = _("Small Capitals")
    tooltip = _("Small Capitals")
    icon_name = "sc"
    snippet_source = "\\textsc{$0}"


class LaTeXRomanAction(LaTeXTemplateAction):
    label = _("Roman")
    tooltip = _("Roman")
    icon_name = "rm"
    snippet_source = "\\textrm{$0}"


class LaTeXSansSerifAction(LaTeXTemplateAction):
    label = _("Sans Serif")
    tooltip = _("Sans Serif")
    icon_name = "sf"
    snippet_source = "\\textsf{$0}"


class LaTeXTypewriterAction(LaTeXTemplateAction):
    label = _("Typewriter")
    tooltip = _("Typewriter")
    icon_name = "tt"
    snippet_source = "\\texttt{$0}"


class LaTeXBlackboardBoldAction(LaTeXTemplateAction):
    label = _("Blackboard Bold")
    tooltip = _("Blackboard Bold")
    icon_name = "bb"
    packages = ["amsmath"]
    snippet_source = "\ensuremath{\mathbb{$0}}"


class LaTeXCaligraphyAction(LaTeXTemplateAction):
    label = _("Calligraphy")
    tooltip = _("Calligraphy")
    icon_name = "cal"
    snippet_source = "\ensuremath{\mathcal{$0}}"


class LaTeXFrakturAction(LaTeXTemplateAction):
    label = _("Fraktur")
    tooltip = _("Fraktur")
    icon_name = "frak"
    packages = ["amsmath"]
    snippet_source = "\ensuremath{\mathfrak{$0}}"


class LaTeXListMenuAction(LaTeXMenuAction):
    label = _("List")
    tooltip = _("List")


class LaTeXItemizeAction(LaTeXTemplateAction):
    label = _("Itemize")
    tooltip = _("Itemize")
    icon_name = "itemize"
    snippet_source = "\\begin{itemize}\n\t\\item $0\n\\end{itemize}"


class LaTeXListActionDefault(LaTeXItemizeAction):
    menu_tool_action = True
    label = _("List")
    tooltip = _("List")


class LaTeXEnumerateAction(LaTeXTemplateAction):
    label = _("Enumerate")
    tooltip = _("Enumerate")
    icon_name = "enumerate"
    snippet_source = "\\begin{enumerate}\n\t\\item $0\n\\end{enumerate}"


class LaTeXDescriptionAction(LaTeXTemplateAction):
    label = _("Description")
    tooltip = _("Description")
    icon_name = "description"
    snippet_source = "\\begin{description}\n\t\\item[$0]\n\\end{description}"

class LaTeXStructureMenuAction(LaTeXMenuAction):
    label = _("Structure")
    tooltip = _("Structure")

class LaTeXPartAction(LaTeXTemplateAction):
    label = _("Part")
    tooltip = _("Part")
    icon_name = "part"
    snippet_source = "\\part{$0}"


class LaTeXChapterAction(LaTeXTemplateAction):
    label = _("Chapter")
    tooltip = _("Chapter")
    icon_name = "chapter"
    snippet_source = "\\chapter{$0}"


class LaTeXSectionAction(LaTeXTemplateAction):
    label = _("Section")
    tooltip = _("Section")
    icon_name = "section"
    snippet_source = "\\section{$0}"


class LaTeXStructureActionDefault(LaTeXSectionAction):
    menu_tool_action = True
    label = _("Structure")
    tooltip = _("Structure")


class LaTeXSubsectionAction(LaTeXTemplateAction):
    label = _("Subsection")
    tooltip = _("Subsection")
    icon_name = "subsection"
    snippet_source = "\\subsection{$0}"


class LaTeXParagraphAction(LaTeXTemplateAction):
    label = _("Paragraph")
    tooltip = _("Paragraph")
    icon_name = "paragraph"
    snippet_source = "\\paragraph{$0}"


class LaTeXSubparagraphAction(LaTeXTemplateAction):
    label = _("Subparagraph")
    tooltip = _("Subparagraph")
    icon_name = "paragraph"
    snippet_source = "\\subparagraph{$0}"


class LaTeXGraphicsAction(LaTeXIconAction):
    label = _("Insert Graphics")
    accelerator = None
    tooltip = _("Insert Graphics")
    icon_name = "graphics"

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = InsertGraphicsDialog()
        source = self.dialog.run(context.active_editor.edited_file)
        if source:
            context.active_editor.insert(source)


class LaTeXTableAction(LaTeXIconAction):
    label = _("Insert Table or Matrix")
    accelerator = None
    tooltip = _("Insert Table or Matrix")
    icon_name = "table"

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = InsertTableDialog()
        source = self.dialog.run()
        if source:
            context.active_editor.insert(source)


class LaTeXListingAction(LaTeXIconAction):
    label = _("Insert Source Code Listing")
    accelerator = None
    tooltip = _("Insert Source Code Listing")
    icon_name = "listing"

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = InsertListingDialog()
        source = self.dialog.run(context.active_editor.edited_file)
        if source:
            context.active_editor.insert(source)

class LaTeXBuildAction(LaTeXIconAction):
    menu_tool_action = True

    label = _("Build")
    accelerator = None
    tooltip = _("Build")
    stock_id = Gtk.STOCK_CONVERT

    def activate(self, context):
        pass


class LaTeXBuildMenuAction(LaTeXMenuAction):
    label = _("Build")
    tooltip = _("Build")


class LaTeXBuildImageAction(LaTeXIconAction):
    label = _("Build Image")
    accelerator = None
    tooltip = _("Build an image from the LaTeX document")
    icon_name = "build-image"

    dialog = None

    def activate(self, context):
        if not self.dialog:
            self.dialog = BuildImageDialog()

        tool = self.dialog.run()
        if tool is not None:
            tool_view = context.find_view(None, "ToolView")

            if context.active_editor:
                ToolRunner().run(context.active_editor.file, tool, tool_view)


class LaTeXJustifyMenuAction(LaTeXMenuAction):
    label = _("Justify")
    tooltip = _("Justify")


class LaTeXJustifyLeftAction(LaTeXTemplateAction):
    label = _("Justify Left")
    tooltip = _("Justify Left")
    icon_name = "justify-left"
    snippet_source = "\\begin{flushleft}$0\\end{flushleft}"


class LaTeXJustifyCenterAction(LaTeXTemplateAction):
    label = _("Justify Center")
    tooltip = _("Justify Center")
    icon_name = "justify-center"
    snippet_source = "\\begin{center}$0\\end{center}"


class LaTeXJustifyActionDefault(LaTeXJustifyCenterAction):
    menu_tool_action = True
    label = _("Justify")
    tooltip = _("Justify")


class LaTeXJustifyRightAction(LaTeXTemplateAction):
    label = _("Justify Right")
    tooltip = _("Justify Right")
    icon_name = "justify-right"
    snippet_source = "\\begin{flushright}$0\\end{flushright}"


class LaTeXMathMenuAction(LaTeXMenuAction):
    label = _("Math")
    tooltip = _("Math")


class LaTeXMathAction(LaTeXTemplateAction):
    label = _("Mathematical Environment")
    tooltip = _("Mathematical Environment")
    icon_name = "math"
    snippet_source = "$ $0 $"


class LaTeXMathActionDefault(LaTeXMathAction):
    menu_tool_action = True
    label = _("Math")
    tooltip = _("Math")


class LaTeXDisplayMathAction(LaTeXTemplateAction):
    label = _("Centered Formula")
    tooltip = _("Centered Formula")
    icon_name = "displaymath"
    snippet_source = "\\[ $0 \\]"


class LaTeXEquationAction(LaTeXTemplateAction):
    label = _("Numbered Equation")
    tooltip = _("Numbered Equation")
    icon_name = "equation"
    snippet_source = """\\begin{equation}
    $0
\\end{equation}"""


class LaTeXUnEqnArrayAction(LaTeXTemplateAction):
    label = _("Array of Equations")
    tooltip = _("Array of Equations")
    icon_name = "uneqnarray"
    packages = ["amsmath"]
    snippet_source = """\\begin{align*}
    $0
\\end{align*}"""


class LaTeXEqnArrayAction(LaTeXTemplateAction):
    label = _("Numbered Array of Equations")
    tooltip = _("Numbered Array of Equations")
    icon_name = "eqnarray"
    packages = ["amsmath"]
    snippet_source = """\\begin{align}
    $0
\\end{align}"""

class FileDummyAction(Action):
    label = tooltip = stock_id = ''
    accelerator = None
    extensions = []

class ToolsDummyAction(Action):
    label = tooltip = stock_id = ''
    accelerator = None
    extensions = []

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
util

Utility classes, functions and decorators used at various places across
the project
"""

import logging

def singleton(cls):
    """
    Singleton decorator that works with GObject derived types. The 'recommended'
    python one - http://wiki.python.org/moin/PythonDecoratorLibrary#Singleton
    does not (interacts badly with GObjectMeta
    """
    instances = {}
    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance

from gi.repository import Gtk
import traceback
from xml.sax import saxutils


class StringReader(object):
    """
    A simple string reader that is able to push back one character
    """
    def __init__(self, string):
        self._iter = iter(string)
        self.offset = 0
        self._pushbackChar = None
        self._pushbackFlag = False

    def read(self):
        if self._pushbackFlag:
            self._pushbackFlag = False
            return self._pushbackChar
        else:
            self.offset += 1
            return next(self._iter)

    def unread(self, char):
        #assert not self._pushbackFlag

        self._pushbackChar = char
        self._pushbackFlag = True


def escape(string, remove_newlines=True):
    """
    Prepares a string for inclusion in Pango markup and error messages
    """
    s = saxutils.escape(string)
    s = s.replace("\n", " ")
    s = s.replace("\"", "&quot;")
    return s


def open_error(message, secondary_message=None):
    """
    Popup an error dialog window
    """
    dialog = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                            Gtk.MessageType.ERROR, Gtk.ButtonsType.OK, message)
    if secondary_message:
        # TODO: why not use markup?
        dialog.format_secondary_text(secondary_message)
    dialog.run()
    dialog.destroy()


def open_info(message, secondary_message=None):
    """
    Popup an info dialog window
    """
    dialog = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
                            Gtk.MessageType.INFO, Gtk.ButtonsType.OK, message)
    if secondary_message:
        dialog.format_secondary_markup(secondary_message)
    dialog.run()
    dialog.destroy()


def verbose(function):
    """
    'verbose'-decorator. This runs the decorated method in a try-except-block
    and shows an error dialog on exception.
    """
    def decorated_function(*args, **kw):
        try:
            return function(*args, **kw)
        except Exception as e:
            stack = traceback.format_exc(limit=10)
            open_error(str(e), stack)
    return decorated_function


class GladeInterface(object):
    """
    Utility base class for interfaces loaded from a Glade definition
    """

    __log = logging.getLogger("GladeInterface")

    def __init__(self):
        self.__tree = None
        self.filename = None

    def __get_tree(self):
        if not self.__tree:
            self.__tree = Gtk.Builder()
            self.__tree.set_translation_domain('gedit-latex')
            self.__tree.add_from_file(self.filename)
        return self.__tree

    def find_widget(self, name):
        """
        Find a widget by its name
        """
        widget = self.__get_tree().get_object(name)
        if widget is None:
            self.__log.error("Widget '%s' could not be found in interface description '%s'" % (name, self.filename))
        return widget

    def connect_signals(self, mapping):
        """
        Auto-connect signals
        """
        self.__get_tree().connect_signals(mapping)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.expander
"""

from logging import getLogger

from ..file import File
from .cache import LaTeXDocumentCache
from .parser import Node


class LaTeXReferenceExpander(object):
    """
    This expands '\include' and '\input' commands by parsing the referenced child
    documents. The resulting trees may be attached to the parent tree.
    """

    # TODO: embed this into parser so that we don't need to walk the document again

    _log = getLogger("ReferenceExpander")

    def __init__(self):
        self._document_cache = LaTeXDocumentCache()

    def expand(self, documentNode, master_file, issue_handler, charset):
        """
        @param documentNode: the master model
        @param master_file: the File object of the master
        @param issue_handler: an IIssueHandler object
        @param charset: a string naming the character set used by Gedit
        """
        self._master_file = master_file
        self._issue_handler = issue_handler
        self._charset = charset
        self._expand(documentNode)

    def _expand(self, parentNode):
        """
        Recursively walk all nodes in the document model and check for \input or \include
        commands. Extract the filename from the command and parse the referenced file.
        Attach its model as a DOCUMENT node to the master model and hold its filename as
        the value of that DOCUMENT node, so that the Validator may differ between issue
        sources.
        """
        for node in parentNode:
            if node.type == Node.COMMAND and (node.value == "input" or node.value == "include"):
                try:
                    # build child filename (absolute/relative, with .tex/without .tex)
                    target = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText

                    file = None
                    if File.is_absolute(target):
                        # absolute path
                        # look for 'x' and then for 'x.tex'
                        file = File("%s.tex" % target)
                        if not file.exists:
                            file = File(target)
                    else:
                        # path relative to the master file's directory
                        # TODO: include TeX search path!
                        # look for 'x' and then for 'x.tex'
                        file = File.create_from_relative_path("%s.tex" % target, self._master_file.dirname)
                        if not file.exists:
                            file = File.create_from_relative_path(target, self._master_file.dirname)

                    self._log.debug("Expanding %s" % file)

                    # lookup/parse child document model
                    try:
                        fragment = self._document_cache.get_document(file, self._charset, self._issue_handler)

                        node.append(fragment)
                    except IOError:
                        self._log.error("Referenced file not found: %s" % file.uri)
                except IndexError:
                    self._log.error("Malformed reference command at %s" % node.start)

            self._expand(node)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

import gettext

VERSION = "3.20.0"
PACKAGE = "gedit-latex"
PACKAGE_STRING = "gedit-latex 3.20.0"
GETTEXT_PACKAGE = "gedit-latex"
GL_LOCALEDIR = "/usr/share/locale"

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GL_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s);
except:
    _ = lambda s: s



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base.job
"""

import dbus
import dbus.service
import dbus.glib
import multiprocessing
import logging


BUS_NAME = 'org.gedit.LaTeXPlugin.JobManager'
OBJECT_PATH = '/org/gedit/LaTeXPlugin/JobManager'


class Job(object):

    __log = logging.getLogger("Job")

    class NoneReturned(object):
        pass

    def __init__(self, argument=None):
        """
        @param arguments: a list of objects to be passed to the Job
        """
        self.__argument = argument
        self.__returned = self.NoneReturned()
        self.__change_listener = None

    def set_argument(self, argument):
        self.__argument = argument

    def schedule(self):
        """
        Run the Job as a subprocess
        """
        # create queue for communication
        self.__queue = multiprocessing.Queue()

        # run process
        self.__process = multiprocessing.Process(target=self.__start, args=(self.__queue,))

        # enqueue argument object
        self.__queue.put(self.__argument)

        # start process
        self.__process.start()

    def abort(self):
        """
        Abort the Job process
        """
        self.__process.terminate()

        # TODO: cleanup?

    def get_returned(self):
        """
        Get the objects returned by the Job
        """
        if type(self.__returned) is self.NoneReturned:
            # dequeue returned object
            self.__returned = self.__queue.get()
        return self.__returned

    def get_exception(self):
        return self.__exception

    @property
    def id(self):
        return id(self)

    def set_change_listener(self, job_change_listener):
        self.__change_listener = job_change_listener

    def __start(self, queue):
        """
        This is started as a subprocess in a separate address space
        """
        # register state change listener
        if not self.__change_listener is None:
            job_manager.add_listener(self.id, self.__change_listener)

        # notify state change
        job_manager.change_state(self.id, JobManager.STATE_STARTED)

        # dequeue argument object
        argument = queue.get()

        # run the job
        self.__exception = None
        try:
            returned = self._run(argument)
        except Exception as e:
            self.__log.error(e)
            self.__exception = e

        # enqueue returned object
        queue.put(returned)

        # notify state change
        job_manager.change_state(self.id, JobManager.STATE_COMPLETED)

        # deregister state change listener
        if not self.__change_listener is None:
            job_manager.remove_listener(self.id)

    def _run(self, arguments):
        """
        @return: a list of objects that should be made available after completion
        """
        pass


class JobChangeListener(object):
    """
    Callback oject for listening to the state changes of a Job
    """
    def _on_state_changed(self, state):
        pass


class GlobalJobChangeListener(object):
    """
    Callback oject for listening to the state changes of ALL Jobs
    """
    def _on_state_changed(self, job_id, state):
        pass


class JobManager(dbus.service.Object):

    STATE_STARTED, STATE_COMPLETED = 1, 2

    __log = logging.getLogger("JobManager")

    def __init__(self):
        bus_name = dbus.service.BusName(BUS_NAME, bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, OBJECT_PATH)

        self.__global_listener = None
        self.__listeners = {}

        self.__log.debug("Created JobManager instance %s" % id(self))

    @dbus.service.method(dbus_interface="org.gedit.JobManagerInterface")
    def change_state(self, job_id, state):
        """
        The job with id <job_id> has changed its state to <state>
        """
        self.__log.debug("change_state(%s, %s)" % (job_id, state))

        # notify global listener
        if self.__global_listener is None:
            self.__log.warn("No global listener")
        else:
            self.__global_listener._on_state_changed(job_id, state)

        # notify listener if present
        try:
            self.__listeners[job_id]._on_state_changed(state)
        except KeyError:
            self.__log.warn("No listener for job %s" % job_id)

    def set_global_listener(self, global_job_change_listener):
        self.__global_listener = global_job_change_listener

    def add_listener(self, job_id, job_change_listener):
        self.__listeners[job_id] = job_change_listener

    def remove_listener(self, job_id):
        del self.__listeners[job_id]

    def dispose(self):
        """
        End life-cycle
        """
        self.__log.debug("dispose")


job_manager = JobManager()


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
preferences
"""

from gi.repository import GObject, Gio, GLib

import re
import os.path
import logging
import configparser

from ..util import singleton

LOG = logging.getLogger(__name__)

class _DocumentConfigParser(configparser.RawConfigParser):

    SECTION = "LATEX"

    def __init__(self, filename):
        configparser.RawConfigParser.__init__(self)
        self._filename = filename
        self.read(filename)
        try:
            self.add_section(self.SECTION)
        except configparser.DuplicateSectionError:
            pass

    def __getitem__(self, key):
        try:
            return configparser.RawConfigParser.get(self,self.SECTION, key)
        except configparser.NoOptionError:
            raise KeyError

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def set(self, key, value):
        configparser.RawConfigParser.set(self, self.SECTION, key, value)

    def save(self):
        f = open(self._filename,'w')
        self.write(f)
        f.close()

class _Preferences(GObject.GObject):

    __gsignals__ = {
        "preferences-changed": (
            GObject.SignalFlags.RUN_LAST, None, [str, str]),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

    def get(self, key, default=None):
        raise NotImplementedError

    def set(self, key, value):
        pass

@singleton
class Preferences(_Preferences):
    """
    A simple map storing preferences as key-value-pairs
    """

    TEMPLATE_DIR = os.path.join(GLib.get_user_data_dir(), "gedit", "latex", "templates")

    def __init__(self):
        _Preferences.__init__(self)
        self._settings = Gio.Settings("org.gnome.gedit.plugins.latex")
        
        LOG.debug("Pref singleton constructed")

    def get(self, key):
        LOG.debug("Get pref: %s" % key)
        return self._settings[key]

    def set(self, key, value):
        LOG.debug("Set pref: %s = %s" % (key,value))
        self._settings[key] = value
        self.emit("preferences-changed", str(key), str(value))

class DocumentPreferences(_Preferences):
    """
    Similar to @Preferences, but first tries to **GET** keys from the current
    document. Searches for lines of the following

    % gedit:key-name = value

    If that fails, it also looks in a .filename.ini file for key-name = value
    lines. If that fails, look in the system settings.

    When **SETTING** keys, they do not persist if they were previously defined in
    the document text, otherwise, they are set in the .ini file.
    """

    def __init__(self, file):
        _Preferences.__init__(self)
        self._sysprefs = Preferences()
        self._sysprefs.connect("preferences-changed", self._on_prefs_changed)
        self._file = file
        self._cp = _DocumentConfigParser(
                        "%s/.%s.ini" % (file.dirname, file.basename))

        self._re = re.compile("^\s*%+\s*gedit:(.*)\s*=\s*(.*)")
        self._modelines = {}

        LOG.debug("Document preferences for %s" % file.basename)

    def _on_prefs_changed(self, p, key, value):
        self.emit("preferences-changed", key, value)

    def parse_content(self, content, max_lines=100):
        """ Parses txt content from the document looking for modelines """
        self._modelines = {}

        i = 0
        for l in content.splitlines():
            if i > max_lines:
                break
            try:
                key,val = self._re.match(l).groups()
                LOG.debug("Document %s prefs modeline: %s = %s" % (self._file.basename,key,val))
                self._modelines[key.strip()] = val
            except AttributeError:
                pass
            i = i+1

    def get(self, key):
        try:
            val = self._modelines[key]
            method = "modeline"
        except KeyError:
            try:
                val = self._cp[key]
                method = "configfile"
            except KeyError:
                try:
                    val = self._sysprefs.get(key)
                    method = "system"
                except KeyError:
                    val = None
                    method = "none"

        LOG.debug("Get doc %s pref: %s = %s (from: %s)" % (self._file.basename,key,val,method))

        return val

    def set(self, key, value):
        try:
            self._sysprefs.set(key, value)
            method = "system"
        except KeyError:
            self._cp.set(key,value)
            self._cp.save()
            self.emit("preferences-changed", key, value)
            method = "configfile"

        LOG.debug("Set doc %s pref: %s = %s (from: %s)" % (self._file.basename,key,value,method))
            

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
views
"""

from gi.repository import Gtk, GdkPixbuf
from logging import getLogger

from .preferences import Preferences
from .resources import Resources
from .panelview import PanelView
from .issues import Issue
from .util import escape
from .gldefs import _


class IssueView(PanelView):
    """
    """

    _log = getLogger("IssueView")

    def __init__(self, context, editor):
        PanelView.__init__(self, context)
        self._log.debug("init")

        self._editor = editor
        self._handlers = {}
        self._preferences = Preferences()

        self._preferences.connect("preferences-changed", self._on_preferences_changed)
        self._show_tasks = self._preferences.get("issues-show-tasks")
        self._show_warnings = self._preferences.get("issues-show-warnings")

        self._icons = { Issue.SEVERITY_WARNING : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("warning.png")),
                        Issue.SEVERITY_ERROR : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("error.png")),
                        Issue.SEVERITY_INFO : None,
                        Issue.SEVERITY_TASK : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("task.png")) }

        grid = Gtk.Grid()
        self.add(grid)

        self._store = Gtk.ListStore(GdkPixbuf.Pixbuf, str, str, object)

        self._view = Gtk.TreeView(model=self._store)

        column = Gtk.TreeViewColumn()
        column.set_title(_("Message"))

        pixbuf_renderer = Gtk.CellRendererPixbuf()
        column.pack_start(pixbuf_renderer, False)
        column.add_attribute(pixbuf_renderer, "pixbuf", 0)

        text_renderer = Gtk.CellRendererText()
        column.pack_start(text_renderer, True)
        column.add_attribute(text_renderer, "markup", 1)

        self._view.append_column(column)

        column = Gtk.TreeViewColumn()
        column.set_title(_("File"))
        text_renderer2 = Gtk.CellRendererText()
        column.pack_start(text_renderer2, True)
        column.add_attribute(text_renderer2, "markup", 2)
        self._view.insert_column(column, -1)
        self._handlers[self._view] = self._view.connect("row-activated", self._on_row_activated)

        self._scr = Gtk.ScrolledWindow()

        self._scr.add(self._view)
        self._scr.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self._scr.set_shadow_type(Gtk.ShadowType.IN)
        self._scr.set_hexpand(True)
        self._scr.set_vexpand(True)

        grid.add(self._scr)

        # toolbar
        self._button_warnings = Gtk.ToggleToolButton()
        self._button_warnings.set_tooltip_text(_("Show/Hide Warnings"))
        image = Gtk.Image()
        image.set_from_file(Resources().get_icon("warning.png"))
        self._button_warnings.set_icon_widget(image)
        self._button_warnings.set_active(self._show_warnings)
        self._handlers[self._button_warnings] = self._button_warnings.connect("toggled", self.__on_warnings_toggled)

        self._button_tasks = Gtk.ToggleToolButton()
        self._button_tasks.set_tooltip_text(_("Show/Hide Tasks"))
        imageTask = Gtk.Image()
        imageTask.set_from_file(Resources().get_icon("task.png"))
        self._button_tasks.set_icon_widget(imageTask)
        self._button_tasks.set_active(self._show_tasks)
        self._handlers[self._button_tasks] = self._button_tasks.connect("toggled", self.__on_tasks_toggled)

        toolbar = Gtk.Toolbar()
        toolbar.set_orientation(Gtk.Orientation.VERTICAL)
        toolbar.set_style(Gtk.ToolbarStyle.ICONS)
        toolbar.set_icon_size(Gtk.IconSize.MENU)
        toolbar.insert(self._button_warnings, -1)
        toolbar.insert(self._button_tasks, -1)
        toolbar.set_vexpand(True)

        grid.add(toolbar)

        # theme like gtk3
        ctx = self._scr.get_style_context()
        ctx.set_junction_sides(Gtk.JunctionSides.RIGHT)

        ctx = toolbar.get_style_context()
        ctx.set_junction_sides(Gtk.JunctionSides.LEFT | Gtk.JunctionSides.RIGHT)
        ctx.add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)

        self._issues = []

        self.show_all()

        self._log.debug("init finished")

    def get_label(self):
        return _("Issues")

    def get_icon(self):
        return Gtk.Image.new_from_stock(Gtk.STOCK_DIALOG_INFO, Gtk.IconSize.MENU)

    def _on_row_activated(self, view, path, column):
        """
        A row has been double-clicked on
        """
        issue = self._store.get(self._store.get_iter(path), 3)[0]

        self._context.activate_editor(issue.file)

        #~ # FIXME: this doesn't work correctly
        #~ if not self._context.active_editor is None:
            #~ self._context.active_editor.select(issue.start, issue.end)
        self._editor.select(issue.start, issue.end)

    def _on_preferences_changed(self, prefs, key, value):
        if key == "issues-show-warnings" or key == "issues-show-tasks":
            # update filter
            self._store.clear()
            for issue, local in self._issues:
                self._append_issue_filtered(issue, local)

    def __on_tasks_toggled(self, togglebutton):
        self._show_tasks = togglebutton.get_active()
        self._preferences.set("issues-show-tasks", self._show_tasks)

    def __on_warnings_toggled(self, togglebutton):
        self._show_warnings = togglebutton.get_active()
        self._preferences.set("issues-show-warnings", self._show_warnings)

    def clear(self):
        """
        Remove all issues from the view
        """
        self._store.clear()
        self._issues = []

    def append_issue(self, issue, local=True):
        """
        Append a new Issue to the view

        @param issue: the Issue object
        @param local: indicates whether the Issue occured in the edited file or not
        """
        self._issues.append((issue, local))
        self._append_issue_filtered(issue, local)

    def _append_issue_filtered(self, issue, local):
        if issue.severity == Issue.SEVERITY_WARNING:
            if self._show_warnings:
                self._do_append_issue(issue, local)
        elif issue.severity == Issue.SEVERITY_TASK:
            if self._show_tasks:
                self._do_append_issue(issue, local)
        else:
            self._do_append_issue(issue, local)

    def _do_append_issue(self, issue, local):
        if local:
            message = issue.message
            filename = escape(issue.file.basename)
        else:
            message = "<span color='%s'>%s</span>" % (self._preferences.get("light-foreground-color"), issue.message)
            filename = "<span color='%s'>%s</span>" % (self._preferences.get("light-foreground-color"), issue.file.basename)
        self._store.append([self._icons[issue.severity], message, filename, issue])

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.actions
"""

from ..action import Action
from ..gldefs import _
from .dialogs import InsertBibTeXEntryDialog


class BibTeXMenuAction(Action):
    extensions = [".bib"]
    label = _("BibTeX")
    stock_id = None
    accelerator = None
    tooltip = None

    def activate(self, context):
        pass


class BibTeXNewEntryAction(Action):
    extensions = [".bib"]
    label = _("New BibTeX Entry...")
    stock_id = None
    accelerator = None
    tooltip = _("Create a new BibTeX entry")

    _dialog = None

    def activate(self, context):
        if not self._dialog:
            self._dialog = InsertBibTeXEntryDialog()

        source = self._dialog.run()
        if not source is None:
            context.active_editor.append(source)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base.config
"""

# actions
from .latex.actions import LaTeXMenuAction, LaTeXNewAction, LaTeXChooseMasterAction, \
        LaTeXListMenuAction, LaTeXListActionDefault, \
        LaTeXItemizeAction, LaTeXEnumerateAction, LaTeXFontFamilyAction, LaTeXFontFamilyMenuAction, LaTeXBoldAction, \
        LaTeXItalicAction, LaTeXEmphasizeAction, LaTeXDescriptionAction, LaTeXStructureMenuAction, LaTeXPartAction, LaTeXChapterAction, \
        LaTeXSectionAction, LaTeXSubsectionAction, LaTeXParagraphAction, LaTeXSubparagraphAction, LaTeXStructureActionDefault, \
        LaTeXGraphicsAction, LaTeXUseBibliographyAction, LaTeXTableAction, LaTeXListingAction, LaTeXJustifyLeftAction, \
        LaTeXJustifyMenuAction, LaTeXJustifyActionDefault, \
        LaTeXJustifyCenterAction, LaTeXJustifyRightAction, LaTeXMathMenuAction, LaTeXMathActionDefault, LaTeXMathAction, LaTeXDisplayMathAction, \
        LaTeXEquationAction, LaTeXUnEqnArrayAction, LaTeXEqnArrayAction, LaTeXUnderlineAction, LaTeXSmallCapitalsAction, \
        LaTeXRomanAction, LaTeXSansSerifAction, LaTeXTypewriterAction, LaTeXCloseEnvironmentAction, LaTeXBlackboardBoldAction, \
        LaTeXCaligraphyAction, LaTeXFrakturAction, LaTeXBuildImageAction, \
        LaTeXBuildAction, LaTeXBuildMenuAction, FileDummyAction, ToolsDummyAction

from .bibtex.actions import BibTeXMenuAction, BibTeXNewEntryAction

ACTIONS = [LaTeXMenuAction, LaTeXNewAction, LaTeXChooseMasterAction,
        LaTeXListMenuAction, LaTeXListActionDefault,
        LaTeXItemizeAction, LaTeXEnumerateAction, LaTeXFontFamilyAction, LaTeXFontFamilyMenuAction, LaTeXBoldAction,
        LaTeXItalicAction, LaTeXEmphasizeAction, LaTeXDescriptionAction, LaTeXStructureMenuAction, LaTeXPartAction, LaTeXChapterAction,
        LaTeXSectionAction, LaTeXSubsectionAction, LaTeXParagraphAction, LaTeXSubparagraphAction, LaTeXStructureActionDefault,
        LaTeXGraphicsAction, LaTeXUseBibliographyAction, LaTeXTableAction, LaTeXListingAction, LaTeXJustifyLeftAction,
        LaTeXJustifyMenuAction, LaTeXJustifyActionDefault,
        LaTeXJustifyCenterAction, LaTeXJustifyRightAction, LaTeXMathMenuAction, LaTeXMathActionDefault, LaTeXMathAction, LaTeXDisplayMathAction,
        LaTeXEquationAction, LaTeXUnEqnArrayAction, LaTeXEqnArrayAction, LaTeXUnderlineAction, LaTeXSmallCapitalsAction,
        LaTeXRomanAction, LaTeXSansSerifAction, LaTeXTypewriterAction, LaTeXCloseEnvironmentAction, LaTeXBlackboardBoldAction,
        LaTeXCaligraphyAction, LaTeXFrakturAction, LaTeXBuildImageAction,
        LaTeXBuildAction, LaTeXBuildMenuAction,
        BibTeXMenuAction, BibTeXNewEntryAction,
        FileDummyAction, ToolsDummyAction]

MENUACTIONS = [LaTeXNewAction, LaTeXChooseMasterAction, LaTeXCloseEnvironmentAction, BibTeXNewEntryAction] 

# views
from .views import IssueView
from .latex.views import LaTeXSymbolMapView, LaTeXOutlineView
from .bibtex.views import BibTeXOutlineView
from .tools.views import ToolView

from .preferences import Preferences
LATEX_EXTENSIONS = Preferences().get("latex-extensions").split(",")
BIBTEX_EXTENSIONS = [".bib"]

EDITOR_VIEWS = {}

for e in LATEX_EXTENSIONS:
    EDITOR_VIEWS[e] = {"ToolView": ToolView, "IssueView": IssueView, "LaTeXOutlineView": LaTeXOutlineView, "LaTeXSymbolMapView": LaTeXSymbolMapView}

for e in BIBTEX_EXTENSIONS:
    EDITOR_VIEWS[e] = {"ToolView": ToolView, "IssueView": IssueView, "BibTeXOutlineView": BibTeXOutlineView}


# editors
from .latex.editor import LaTeXEditor
from .bibtex.editor import BibTeXEditor

EDITORS = [LaTeXEditor, BibTeXEditor]


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
issues
"""

from logging import getLogger


class IIssueHandler(object):
    """
    A class implementing this interface handles issues
    """

    def clear(self):
        """
        Remove all partitions and issues
        """
        raise NotImplementedError

    def issue(self, issue):
        """
        An issue has occured
        """
        raise NotImplementedError


class MockIssueHandler(IIssueHandler):
    """
    This is used by the BibTeXDocumentCache
    """
    __log = getLogger("MockIssueHandler")

    def clear(self):
        pass

    def issue(self, issue):
        self.__log.debug(str(issue))


class IStructuredIssueHandler(object):
    def clear(self):
        """
        Remove all partitions and issues
        """
        raise NotImplementedError

    def add_partition(self, label, state, parent_partition_id):
        """
        Add a new partition

        @param label: a label used in the UI
        @param state: the initial state descriptor for the partition
        @param parent_partition_id: the partition under which this one should be
                created (None for top-level)

        @return: a unique id for the partition
        """
        raise NotImplementedError

    def set_partition_state(self, partition_id, state):
        """
        @param partition_id: a partition id as returned by add_partition
        @param state: any string
        """
        raise NotImplementedError

    def set_abort_enabled(self, enabled, method):
        """
        @param enabled: if True a job may be aborted
        @param method: a method that is may be called to abort a running job
        """
        raise NotImplementedError

    def append_issues(self, partition_id, issues):
        """
        An issue occured

        @param issue: an Issue object
        @param partition: a partition id as returned by add_partition
        """
        raise NotImplementedError


class MockStructuredIssueHandler(IStructuredIssueHandler):
    """
    Used by the PreviewRenderer
    """
    def clear(self):
        pass

    def add_partition(self, label, state, parent_partition_id):
        pass

    def set_partition_state(self, partition_id, state):
        pass

    def append_issues(self, partition_id, issues):
        pass

    def set_abort_enabled(self, enabled, method):
        pass


class Issue(object):
    """
    An issue can be a warning, an error, an info or a task that occures or is
    recognized during parsing and validation of a source file
    """

    SEVERITY_WARNING, SEVERITY_ERROR, SEVERITY_INFO, SEVERITY_TASK = 1, 2, 3, 4

    POSITION_OFFSET, POSITION_LINE = 1, 2

    def __init__(self, message, start, end, file, severity, position_type=POSITION_OFFSET):
        """
        @param message: a str in Pango markup
        @param start: the start offset of the issue
        @param end: the end offset
        @param file: the File object representing the file the issue occured in
        @param severity: one of SEVERITY_*
        """
        self.message = message
        self.start = start
        self.end = end
        self.file = file
        self.severity = severity
        self.position_type = position_type

    def __str__(self):
        return "Issue{'%s', %s, %s, %s, %s}" % (self.message, self.start, self.end, self.file, self.severity)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.views

The BibTeX outline view
"""

from gi.repository import Gtk, GdkPixbuf
from xml.sax.saxutils import escape
from logging import getLogger

from ..outline import OutlineOffsetMap, BaseOutlineView
from ..resources import Resources
from ..preferences import Preferences
from .parser import Entry
from ..gldefs import _


GROUP_NONE, GROUP_TYPE, GROUP_AUTHOR, GROUP_YEAR = 1, 2, 3, 4


class BibTeXOutlineView(BaseOutlineView):
    """
    Special outline view for BibTeX
    """

    _log = getLogger("BibTeXOutlineView")

    def __init__(self, context, editor):
        BaseOutlineView.__init__(self, context, editor)
        self._handlers = {}

        self._grouping = GROUP_NONE

        # add grouping controls to toolbar

        self._item_none = Gtk.RadioMenuItem.new_with_label([], _("No Grouping"))
        self._item_type = Gtk.RadioMenuItem.new_with_label((self._item_none,), _("Group by Type"))
        self._item_author = Gtk.RadioMenuItem.new_with_label((self._item_none,), _("Group by Author"))
        self._item_year = Gtk.RadioMenuItem.new_with_label((self._item_none,), _("Group by Year"))

        self._preferences = Preferences()

        grouping = self._preferences.get("bibtex-outline-grouping")
        if grouping == GROUP_NONE:
            self._item_none.set_active(True)
        elif grouping == GROUP_TYPE:
            self._item_type.set_active(True)
        elif grouping == GROUP_AUTHOR:
            self._item_author.set_active(True)
        elif grouping == GROUP_YEAR:
            self._item_year.set_active(True)

        self._handlers[self._item_none] = self._item_none.connect("toggled", self._on_grouping_toggled)
        self._handlers[self._item_type] = self._item_type.connect("toggled", self._on_grouping_toggled)
        self._handlers[self._item_author] = self._item_author.connect("toggled", self._on_grouping_toggled)
        self._handlers[self._item_year] = self._item_year.connect("toggled", self._on_grouping_toggled)

        menu = Gtk.Menu()
        menu.add(self._item_none)
        menu.add(self._item_type)
        menu.add(self._item_author)
        menu.add(self._item_year)
        menu.show_all()

        tool_button = Gtk.MenuToolButton.new_from_stock(Gtk.STOCK_SORT_DESCENDING)
        tool_button.set_menu(menu)
        tool_button.show()

        self._toolbar.insert(tool_button, -1)

        self.show_all()

    def _on_grouping_toggled(self, toggle_button):
        if self._item_none.get_active():
            self._grouping = GROUP_NONE
        elif self._item_type.get_active():
            self._grouping = GROUP_TYPE
        elif self._item_author.get_active():
            self._grouping = GROUP_AUTHOR
        elif self._item_year.get_active():
            self._grouping = GROUP_YEAR

        if self._outline:
            self._update()

    def _update(self):
        #t = time.clock()
        self._offset_map = OutlineOffsetMap()
        OutlineConverter().convert(self._store, self._outline,
                                   self._offset_map, self._grouping)
        #dt = time.clock() - t
        #self._log.debug("OutlineConverter.convert: %fs" % dt)

    def set_outline(self, outline):
        """
        Display the given model
        """
        self._outline = outline

        self._save_state()
        self._update()
        self._restore_state()

    def _on_node_selected(self, node):
        """
        An outline node has been selected
        """
        if isinstance(node, Entry):
            self._editor.select(node.start, node.end)


class OutlineConverter(object):
    """
    This converts a BibTeX document to a Gtk.TreeStore and realizes the
    grouping feature
    """

    def __init__(self):
        self._ICON_ENTRY = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("document.png"))
        self._ICON_FIELD = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("field.png"))
        self._ICON_AUTHOR = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("users.png"))
        self._ICON_YEAR = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("calendar.png"))
        self._ICON_TYPE = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("documents.png"))

    def convert(self, tree_store, document, offset_map, grouping=GROUP_NONE):
        """
        Convert a BibTeX document model into a Gtk.TreeStore

        @param tree_store: the Gtk.TreeStore to fill
        @param document: the BibTeX document model (bibtex.parser.Document object)
        @param offset_map: the OutlineOffsetMap object to be filled
        @param grouping: the grouping to use: GROUP_NONE|GROUP_TYPE|GROUP_AUTHOR|GROUP_YEAR
        """

        color = Preferences().get("light-foreground-color")

        tree_store.clear()

        if grouping == GROUP_TYPE:
            # group by entry type

            groups = {}        # maps lower case entry type names to lists of entries

            # collect
            for entry in document.entries:
                try:
                    entryList = groups[entry.type]
                    entryList.append(entry)
                except KeyError:
                    groups[entry.type] = [entry]

            # sort by type
            entryTypes = list(groups.keys())
            entryTypes.sort()

            # build tree
            for entryType in entryTypes:
                entries = groups[entryType]

                parentType = tree_store.append(None, ["%s <span color='%s'>%s</span>" % (escape(entryType),
                                                                                         color, len(entries)),
                                                      self._ICON_TYPE, None])

                for entry in entries:
                    parentEntry = tree_store.append(parentType, [escape(entry.key), self._ICON_ENTRY, entry])

                    offset_map.put(entry.start, tree_store.get_path(parentEntry))

                    for field in entry.fields:
                        tree_store.append(parentEntry, ["<span color='%s'>%s</span> %s" % (color, escape(field.name),
                                                                                           field.valueMarkup),
                                                        self._ICON_FIELD, field])

        elif grouping == GROUP_YEAR:
            # group by year

            NO_YEAR_IDENT = _("<i>n/a</i>")

            groups = {}

            # collect
            for entry in document.entries:
                try:
                    year = str(entry.findField("year").valueString)
                except KeyError:
                    # no year, so put this in an extra group
                    year = NO_YEAR_IDENT

                try:
                    entries = groups[year]
                    entries.append(entry)
                except KeyError:
                    groups[year] = [entry]

            # sort by year
            years = list(groups.keys())
            years.sort()

            # build tree
            for year in years:
                entries = groups[year]

                parentYear = tree_store.append(None, ["%s <span color='%s'>%s</span>" % (year, color, len(entries)),
                                                      self._ICON_YEAR, None])

                for entry in entries:
                    parentEntry = tree_store.append(parentYear, ["%s <span color='%s'>%s</span>" % (escape(entry.key),
                                                                                                    color, escape(entry.type)),
                                                                 self._ICON_ENTRY, entry])

                    offset_map.put(entry.start, tree_store.get_path(parentEntry))

                    for field in entry.fields:
                        tree_store.append(parentEntry, ["<span color='%s'>%s</span> %s" % (color, escape(field.name),
                                                                                           field.valueMarkup),
                                                        self._ICON_FIELD, field])

        elif grouping == GROUP_AUTHOR:

            NA_IDENT = _("Unknown Author")

            groups = {}

            # group
            for entry in document.entries:
                # split list of authors
                try:
                    authorValue = str(entry.findField("author").valueString)
                    authors = [a.strip() for a in authorValue.split(" and ")]
                except KeyError:
                    # no year, so put this in an extra group
                    authors = [NA_IDENT]

                # add to group(s)
                for author in authors:
                    try:
                        entries = groups[author]
                        entries.append(entry)
                    except KeyError:
                        groups[author] = [entry]

            # sort
            authors = list(groups.keys())
            authors.sort()

            # build tree
            for author in authors:
                entries = groups[author]

                parent = tree_store.append(None, ["%s <span color='%s'>%s</span>" % (escape(author), color, len(entries)),
                                                  self._ICON_AUTHOR, None])

                for entry in entries:
                    parentEntry = tree_store.append(parent, ["%s <span color='%s'>%s</span>" % (escape(entry.key),
                                                                                                color, escape(entry.type)),
                                                             self._ICON_ENTRY, entry])

                    offset_map.put(entry.start, tree_store.get_path(parentEntry))

                    for field in entry.fields:
                        tree_store.append(parentEntry, ["<span color='%s'>%s</span> %s" % (color, escape(field.name),
                                                                                           field.valueMarkup),
                                                        self._ICON_FIELD, field])

        else:
            # no grouping, display entries and fields in a tree

            for entry in document.entries:
                parent = tree_store.append(None, ["%s <span color='%s'>%s</span>" % (escape(entry.key),
                                                                                     color, escape(entry.type)),
                                                  self._ICON_ENTRY, entry])

                offset_map.put(entry.start, tree_store.get_path(parent))

                for field in entry.fields:
                    tree_store.append(parent, ["<span color='%s'>%s</span> %s" % (color, escape(field.name),
                                                                                  field.valueMarkup),
                                               self._ICON_FIELD, field])

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
tools.postprocess
"""

import logging
import re

from ..issues import Issue
from ..util import escape


class PostProcessor(object):
    """
    The contract for a post-processor
    """

    def process(self, file, stdout, stderr, condition):
        """
        @param file: the File processed by the Tool
        @param stdout: the output written to STDOUT
        @param stderr: the output written to STDERR
        @param condition: the exit condition of the Tool process
        """
        raise NotImplementedError

    @property
    def successful(self):
        """
        Return whether the Tool process was successful
        """
        raise NotImplementedError

    @property
    def issues(self):
        """
        Return a list of Issues
        """
        raise NotImplementedError

    @property
    def summary(self):
        """
        Return a short string summarizing the result of the Tool process
        """
        raise NotImplementedError


class GenericPostProcessor(PostProcessor):
    """
    This just interprets the exit condition of the process
    """

    _log = logging.getLogger(__name__ + ".GenericPostProcessor")

    name = "GenericPostProcessor"

    def __init__(self):
        self._issues = None
        self._summary = None

    def process(self, file, stdout, stderr, condition):
        self._issues = []
        self._summary = ""
        self._condition = condition

    @property
    def successful(self):
        return not bool(self._condition)

    @property
    def issues(self):
        return self._issues

    @property
    def summary(self):
        return self._summary


class LaTeXPostProcessor(PostProcessor):
    """
    This post-processor generates messages from a standard LaTeX log with
    default error format (NOT using "-file-line-error")
    """

    _log = logging.getLogger(__name__ + ".LatexPostProcessor")

    name = "LaTeXPostProcessor"

    _PATTERN = pattern = re.compile(r"(^! (?P<text>.*?)$)|(^l\.(?P<line>[0-9]+))", re.MULTILINE)

    def __init__(self):
        self._successful = False

    def process(self, file, stdout, stderr, condition):
        self._file = file
        self._successful = not bool(condition)

    @property
    def successful(self):
        return self._successful

    @property
    def summary(self):
        return self._summary

    @property
    def issues(self):
        try:
            log = open("%s.log" % self._file.shortname).read()

            # check for wrong format
            if log.find("file:line:error") > 0:
                return [Issue("The file:line:error format is not supported. Please remove that switch.", None, None, self._file, Issue.SEVERITY_ERROR)]

            # process log file and extract tuples like (message, line)
            tuples = []
            for match in self._PATTERN.finditer(log):
                if match.group("text"):
                    tuple = [match.group("text"), 0]
                elif match.group("line"):
                    tuple[1] = match.group("line")
                    tuples.append(tuple)

            # generate issues from tuples
            self._issues = []
            for tuple in tuples:
                text = tuple[0]
                line = int(tuple[1]) - 1
                self._issues.append(Issue(text, line, None, self._file, Issue.SEVERITY_ERROR, Issue.POSITION_LINE))

            return self._issues

        except IOError:
            return [Issue("No LaTeX log file found", None, None, self._file, Issue.SEVERITY_ERROR)]


class RubberPostProcessor(PostProcessor):
    """
    This is a post-processor for rubber
    """

    _log = logging.getLogger(__name__ + ".RubberPostProcessor")

    name = "RubberPostProcessor"

    _pattern = re.compile(r"(?P<file>[a-zA-Z0-9./_-]+)(:(?P<line>[0-9\-]+))?:(?P<text>.*?)$", re.MULTILINE)

    def __init__(self):
        self._issues = None
        self._summary = None
        self._successful = False

        # FIXME: circ dep
        from ..preferences import Preferences

        self._hide_box_warnings = Preferences().get("hide-box-warnings")

    @property
    def successful(self):
        return self._successful

    @property
    def issues(self):
        return self._issues

    @property
    def summary(self):
        return self._summary

    def process(self, file, stdout, stderr, condition):
        from ..file import File        # FIXME: this produces a circ dep on toplevel

        self._issues = []

        self._log.debug("process(): stdout=\"%s\", stderr=\"%s\"" % (stdout, stderr))

        self._successful = not bool(condition)

        for match in self._pattern.finditer(stderr):
            severity = Issue.SEVERITY_ERROR

            # text
            text = match.group("text")

            if "Underfull" in text or "Overfull" in text:
                if self._hide_box_warnings:
                    continue
                else:
                    severity = Issue.SEVERITY_WARNING

            # line(s)
            lineFrom, lineTo = None, None

            line = match.group("line")

            if line:
                parts = line.split("-")
                lineFrom = int(parts[0]) - 1
                if len(parts) > 1:
                    lineTo = int(parts[1]) - 1

            # filename
            filename = "%s/%s" % (file.dirname, match.group("file"))

            self._issues.append(Issue(escape(text), lineFrom, lineTo, File(filename), severity, Issue.POSITION_LINE))



# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from os import remove
from glob import glob

import logging
import os.path

import re
import urllib.request, urllib.parse, urllib.error

class File(object):
    """
    This is an object-oriented wrapper for all the os.* stuff. A File object
    represents the reference to a file.
    """

    # TODO: use Gio.File as underlying implementation

    @staticmethod
    def create_from_relative_path(relative_path, working_directory):
        """
        Create a File from a path relative to some working directory.

        File.create_from_relative_path('../sub/file.txt', '/home/michael/base') == File('/home/michael/sub/file.txt')

        @param relative_path: a relative path, e.g. '../../dir/myfile.txt'
        @param working_directory: an absolute directory to be used as the starting point for the relative path
        """
        absolute_path = os.path.abspath(os.path.join(working_directory, relative_path))
        return File(absolute_path)

    @staticmethod
    def is_absolute(path):
        return os.path.isabs(path)

    __log = logging.getLogger("File")

    _DEFAULT_SCHEME = "file://"

    def __init__(self, uri):
        """
        @param uri: any URI, URL or local filename
        """
        if uri is None:
            raise ValueError("URI must not be None")

        self._uri = urllib.parse.urlparse(uri)
        if len(self._uri.scheme) == 0:
            # prepend default scheme if missing
            self._uri = urllib.parse.urlparse("%s%s" % (self._DEFAULT_SCHEME, uri))

    def create(self, content=None):
        """
        Create a the File in the file system
        """
        f = open(self.path, "w")
        if content is not None:
            f.write(content)
        f.close()

    @property
    def path(self):
        """
        Returns '/home/user/image.jpg' for 'file:///home/user/image.jpg'
        """
        return urllib.request.url2pathname(self._uri.path)

    @property
    def extension(self):
        """
        Returns '.jpg' for 'file:///home/user/image.jpg'
        """
        return os.path.splitext(self.path)[1]

    @property
    def shortname(self):
        """
        Returns '/home/user/image' for 'file:///home/user/image.jpg'
        """
        return os.path.splitext(self.path)[0]

    @property
    def basename(self):
        """
        Returns 'image.jpg' for 'file:///home/user/image.jpg'
        """
        return os.path.basename(self.path)

    @property
    def shortbasename(self):
        """
        Returns 'image' for 'file:///home/user/image.jpg'
        """
        return os.path.splitext(os.path.basename(self.path))[0]

    @property
    def dirname(self):
        """
        Returns '/home/user' for 'file:///home/user/image.jpg'
        """
        return os.path.dirname(self.path)

    @property
    def uri(self):
        return self._uri.geturl()

    @property
    def exists(self):
        return os.path.exists(self.path)

    @property
    def mtime(self):
        if self.exists:
            return os.path.getmtime(self.path)
        else:
            raise IOError("File not found")

    def find_neighbors(self, extension):
        """
        Find other files in the directory of this one having
        a certain extension

        @param extension: a file extension pattern like '.tex' or '.*'
        """

        # TODO: glob is quite expensive, find a simpler way for this

        try:
            filenames = glob("%s/*%s" % (self.dirname, extension))
            neighbors = [File(filename) for filename in filenames]
            return neighbors

        except Exception as e:
            # as seen in Bug #2002630 the glob() call compiles a regex and so we must be prepared
            # for an exception from that because the shortname may contain regex characters

            # TODO: a more robust solution would be an escape() method for re

            self.__log.debug("find_neighbors: %s" % e)

            return []

    @property
    def siblings(self):
        """
        Find other files in the directory of this one having the same
        basename. This means for a file '/dir/a.doc' this method returns
        [ '/dir/a.tmp', '/dir/a.sh' ]
        """
        siblings = []
        try:
            filenames = glob("%s.*" % self.shortname)
            siblings = [File(filename) for filename in filenames]
        except Exception as e:
            # as seen in Bug #2002630 the glob() call compiles a regex and so we must be prepared
            # for an exception from that because the shortname may contain regex characters

            # TODO: a more robust solution would be an escape() method for re

            self.__log.debug("find_siblings: %s" % e)
        return siblings

    def relativize(self, base, allow_up_level=False):
        """
        Relativize the path of this File against a base directory. That means that e.g.
        File("/home/user/doc.tex").relativize("/home") == "user/doc.tex"

        If up-level references are NOT allowed but necessary (e.g. base='/a/b/c', path='/a/b/d')
        then the absolute path is returned.

        @param base: the base directory to relativize against
        @param allow_up_level: allow up-level references (../../) or not
        """
        if allow_up_level:
            return os.path.relpath(self.path, base)
        else:
            # TODO: why do we need this?

            # relative path must be 'below' base path
            if len(base) >= len(self.path):
                return self.path
            if self.path[:len(base)] == base:
                # bases match, return relative part
                return self.path[len(base) + 1:]
            return self.path

    def relativize_shortname(self, base):
        """
        Relativize the path of this File and return only the shortname of the resulting
        relative path. That means that e.g.
        File("/home/user/doc.tex").relativize_shortname("/home") == "user/doc"

        This is just a convenience method.

        @param base: the base directory to relativize against
        """
        relative_path = self.relativize(base)
        return os.path.splitext(relative_path)[0]

    def delete(self):
        """
        Delete the File from the file system

        @raise OSError:
        """
        if self.exists:
            remove(self.path)
        else:
            raise IOError("File not found")

    def __eq__(self, other):
        """
        Override == operator
        """
        try:
            return self.uri == other.uri
        except AttributeError:        # no File object passed or None
            # returning NotImplemented is bad because we have to
            # compare None with File
            return False

    def __ne__(self, other):
        """
        Override != operator
        """
        return not self.__eq__(other)

    def __str__(self):
        return self.uri

    def __key__(self, file):
        try:
            return file.basename
        except AttributeError:        # no File object passed or None
            # returning NotImplemented is bad because we have to
            # compare None with File
            return None

    def __lt__(self, other):
        return self.__key__(self) < self.__key__(other)

    def __eq__(self, other):
        return self.__key__(self) == self.__key__(other)

    def __hash__(self):
        return hash(self.__key__(self))


class Folder(File):

    # FIXME: a Folder is NOT a subclass of a File, both are a subclass of some AbstractFileSystemObject,
    # this is just a quick hack
    #
    # FIXME: but basically a Folder is a File so this class should not be needed

    __log = logging.getLogger("Folder")

    @property
    def files(self):
        """
        Return File objects for all files in this Folder
        """
        try:
            filenames = glob("%s/*" % (self.path))
            files = [File(filename) for filename in filenames]
            return files

        except Exception as e:
            # as seen in Bug #2002630 the glob() call compiles a regex and so we must be prepared
            # for an exception from that because the shortname may contain regex characters

            # TODO: a more robust solution would be an escape() method for re

            self.__log.debug("files: %s" % e)

            return []

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.model

The LaTeX language model used for code completion.
"""

import copy
import logging

from ..resources import Resources

LOG = logging.getLogger(__name__)

class Element(object):
    """
    """
    TYPE_COMMAND, TYPE_MANDATORY_ARGUMENT, TYPE_OPTIONAL_ARGUMENT, TYPE_CHOICE, TYPE_PLACEHOLDER = 1, 2, 3, 4, 5

    def __init__(self, package, type):
        """
        @param package: the name of the LaTeX package providing this Element
        @param type: one of TYPE_*
        """
        self.package = package
        self.type = type
        self._children = []

    @property
    def children(self):
        return self._children

    def append_child(self, child):
        self._children.append(child)


class Command(Element):
    def __init__(self, package, name):
        Element.__init__(self, package, Element.TYPE_COMMAND)
        self.name = name

    @property
    def first_mandatory_argument(self):
        for node in self.children:
            if node.type == Element.TYPE_MANDATORY_ARGUMENT:
                return node
        raise IndexError

    @property
    def first_optional_argument(self):
        for node in self.children:
            if node.type == Element.TYPE_OPTIONAL_ARGUMENT:
                return node
        raise IndexError


class Argument(Element):
    def get_children(self):
        """
        Override the children property of Element to be able to evaluate Placeholders
        allowed as children of Arguments
        """
        children = []

        for node in self._children:
            if node.type == Element.TYPE_PLACEHOLDER:
                children.extend(node.children)
            else:
                children.append(node)
        return children

    def set_children(self, children):
        self._children = children

    children = property(get_children, set_children)


class MandatoryArgument(Argument):
    def __init__(self, package, label):
        Argument.__init__(self, package, Element.TYPE_MANDATORY_ARGUMENT)
        self.label = label


class OptionalArgument(Argument):
    def __init__(self, package, label):
        Argument.__init__(self, package, Element.TYPE_OPTIONAL_ARGUMENT)
        self.label = label


class Choice(Element):
    def __init__(self, package, value, details=None):
        Element.__init__(self, package, Element.TYPE_CHOICE)
        self.value = value
        self.details = details


class Placeholder(Element):
    def __init__(self, name):
        Element.__init__(self, None, Element.TYPE_PLACEHOLDER)
        self.name = name

    def get_children(self):
        return self._children

    def set_children(self, children):
        self._children = children

    children = property(get_children, set_children)


#LanguageModel is pickled to save startup time...
#using pickle is dangerous. If you add or change the instance members of this class
#that need to persis through pickle, then INCREMENT THE VERSION
LANGUAGE_MODEL_VERSION = 1

class LanguageModel(object):
    """
    """

    REF_CMDS = set(("ref","eqref","pageref"))

    def __init__(self):
        self.commands = {}            # maps command names to Command elements
        self.VERSION = LANGUAGE_MODEL_VERSION

        self.__placeholders = {}
        self.__newcommands = []

        #some latex specific helpers.
        self.__new_ref_commands = {}

    def find_command(self, prefix):
        """
        Find a command by a prefix. A prefix like 'be' would return the command '\begin'
        """
        return [command for name, command in self.commands.items() if name.startswith(prefix)]

    def register_placeholder(self, placeholder):
        """
        Register a placeholder under its name. There may be multiple
        placeholder nodes for one name.
        """
        try:
            nodes = self.__placeholders[placeholder.name]
            nodes.append(placeholder)
        except KeyError:
            self.__placeholders[placeholder.name] = [placeholder]

    def fill_placeholder(self, name, child_elements):
        """
        Attach child elements to a placeholder
        """
        try:
            for placeholder in self.__placeholders[name]:
                placeholder.children = child_elements
        except KeyError:
            LOG.info("fill_placeholder: placeholder '%s' not registered" % name)

    def is_ref_command(self, cmd_name):
        return (cmd_name in self.REF_CMDS) or (cmd_name in self.__new_ref_commands) 

    def set_newcommands(self, outlinenodes):
        LOG.debug("set newcommands")

        #remove old state
        self.__new_ref_commands = {}
        for name in self.__newcommands:
            del(self.commands[name])
        self.__newcommands = []

        for o in outlinenodes:
            #if this is a redefinition of an existing node then use that node as
            #the completion helper
            if o.oldcmd and o.oldcmd in self.commands:
                LOG.info("Detected redefined \\newcommand: %s -> %s" % (o.value, o.oldcmd))
                #copy the old command so we retain its argument completion but change
                #the display name
                old = copy.copy(self.commands[o.oldcmd])
                old.name = o.value

                self.commands[o.value] = old
                self.__newcommands.append(o.value)

                if o.oldcmd in self.REF_CMDS:
                    self.__new_ref_commands[o.value] = 1

            else:
                #add a generic completer
                command = Command(None, o.value)
                for i in range(o.numOfArgs):
                    command.children.append(MandatoryArgument(None, "#%s" % (i + 1)))

                self.commands[command.name] = command
                self.__newcommands.append(command.name)

from xml import sax

class LanguageModelParser(sax.ContentHandler):
    """
    SAX parser for the language model in latex.xml
    """

    # TODO: this should be a simple state machine

    def parse(self, filename, language_model):
        self.__language_model = language_model

        self.__command = None
        self.__argument = None

        LOG.debug("parsing %s" % filename)

        sax.parse(filename, self)

    def startElement(self, name, attrs):
        try:
            package = attrs["package"]
        except KeyError:
            package = None

        if name == "command":
            name = attrs["name"]
            self.__command = Command(package, name)
            self.__language_model.commands[name] = self.__command

        elif name == "argument":
            try:
                label = attrs["label"]
            except KeyError:
                label = ""

            try:
                if attrs["type"] == "optional":
                    self.__argument = OptionalArgument(package, label)
                else:
                    self.__argument = MandatoryArgument(package, label)
            except KeyError:
                self.__argument = MandatoryArgument(package, label)

            self.__command.children.append(self.__argument)

        elif name == "choice":
            choice = Choice(package, attrs["name"])
            self.__argument.append_child(choice)

        elif name == "placeholder":
            placeholder = Placeholder(attrs["key"])
            self.__argument.append_child(placeholder)
            self.__language_model.register_placeholder(placeholder)


import pickle

from ..file import File


class LanguageModelFactory(object):
    """
    This singleton creates LanguageModel instances.

    If a serialized ('pickled') LanguageModel object exists, then a copy
    of this object is returned. Otherwise the XML file must be parsed.
    """

    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):

            pickled_object = self.__find_pickled_object()

            if pickled_object:
                LOG.debug("language model: pickled object loaded")
                self.language_model = pickled_object
            else:
                LOG.debug("language model: no pickled object loaded")
                pkl_filename = Resources().get_user_file("latex.pkl")
                xml_filename = Resources().get_data_file("latex.xml")

                self.language_model = LanguageModel()
                parser = LanguageModelParser()
                parser.parse(xml_filename, self.language_model)

                pickle.dump(self.language_model, open(pkl_filename, 'wb'))
                LOG.info("Pickling language model")

            self._ready = True

    def __find_pickled_object(self):
        pkl_file = File(Resources().get_user_file("latex.pkl"))
        xml_file = File(Resources().get_data_file("latex.xml"))

        if pkl_file.exists:
            if xml_file.mtime > pkl_file.mtime:
                LOG.debug("Pickled object and XML file have different modification times")
            else:
                try:
                    obj = pickle.load(open(pkl_file.path, "rb"))
                    if obj.VERSION == LANGUAGE_MODEL_VERSION:
                        return obj
                    else:
                        LOG.info("Language model obsolete")
                except:
                    LOG.info("Invalid language model", exc_info=True)

        return None

    def get_language_model(self):
        return self.language_model

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from uuid import uuid1
from gi.repository import Gtk, GdkPixbuf, Gio

from .file import File
from .resources import Resources

class GeditLaTeXPlugin_MenuToolAction(Gtk.Action):
    __gtype_name__ = "GeditLaTeXPlugin_MenuToolAction"

    def do_create_tool_item(self):
        return Gtk.MenuToolButton()


class Action(object):
    """
    """

    menu_tool_action = False    # if True a MenuToolAction is created and hooked for this action
                                # instead of Gtk.Action

    extensions = [None]         # a list of file extensions for which this action should be enabled
                                # [None] indicates that this action is to be enabled for all extensions

    def __init__(self, *args, **kwargs):
        pass

    def hook(self, action_group, window_context):
        """
        Create an internal action object (Gtk.Action or MenuToolAction), listen to it and
        hook it in an action group

        @param action_group: a Gtk.ActionGroup object
        @param window_context: a WindowContext object to pass when this action is activated
        """
        if self.menu_tool_action:
            action_clazz = GeditLaTeXPlugin_MenuToolAction
        else:
            action_clazz = Gtk.Action
        self._internal_action = action_clazz(self.__class__.__name__, self.label, self.tooltip, self.stock_id)
        self._handler = self._internal_action.connect("activate", lambda gtk_action, action: action.activate(window_context), self)
        action_group.add_action_with_accel(self._internal_action, self.accelerator)
    
    # Hooks a Gio.SimpleAction to a given window.
    def simplehook(self, window, window_context):
        self._simple_internal_action = Gio.SimpleAction(name=self.__class__.__name__)
        self._simplehandler = self._simple_internal_action.connect("activate", lambda action, param: self.activate(window_context))
        window.add_action(self._simple_internal_action)
        
    @property
    def label(self):
        raise NotImplementedError

    @property
    def stock_id(self):
        raise NotImplementedError

    @property
    def accelerator(self):
        raise NotImplementedError

    @property
    def tooltip(self):
        raise NotImplementedError

    def activate(self, context):
        """
        @param context: the current WindowContext instance
        """
        raise NotImplementedError

    def unhook(self, action_group):
        self._internal_action.disconnect(self._handler)
        action_group.remove_action(self._internal_action)


class IconAction(Action):
    """
    A utility class for creating actions with a custom icon instead of
    a gtk stock id.

    The subclass must provide a field 'icon'.
    """

    icon_name = None   
    __stock_id = None

    def __init__(self, *args, **kwargs):
        self.__icon_factory = kwargs["icon_factory"]
        self.__icon = None

    @property
    def icon(self):
        """
        Return a File object for the icon to use
        """
        if not self.__icon:
            assert(self.icon_name)
            self.__icon = File(Resources().get_icon("%s.png" % self.icon_name)) 
        return self.__icon

    def __init_stock_id(self):
        #
        # generate a new stock id
        #
        self.__stock_id = str(uuid1())
        self.__icon_factory.add(
                self.__stock_id,
                Gtk.IconSet.new_from_pixbuf(
                    GdkPixbuf.Pixbuf.new_from_file(self.icon.path)))

    @property
    def stock_id(self):
        if self.icon:
            if not self.__stock_id:
                self.__init_stock_id()
            return self.__stock_id
        else:
            return None

# ex:ts=4:et:
